#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 3.0 implementation - part 3 (functions)
"""
import decimal
import os
import re
import codecs
import math
import zoneinfo
from collections.abc import Iterator
from copy import copy
from itertools import zip_longest
from typing import cast, Any, Optional, Union, NoReturn
from urllib.parse import urlsplit
from urllib.request import urlopen
from urllib.error import URLError

import elementpath.aliases as ta

from elementpath.exceptions import ElementPathError
from elementpath.tdop import MultiLabel
from elementpath.helpers import Patterns, is_xml_codepoint, node_position
from elementpath.namespaces import get_expanded_name, split_expanded_name, \
    XPATH_FUNCTIONS_NAMESPACE
from elementpath.datatypes import NumericProxy, QName, Date, DateTime, Time, AnyURI
from elementpath.sequences import xlist
from elementpath.sequence_types import is_sequence_type, match_sequence_type
from elementpath.etree import defuse_xml
from elementpath.xpath_nodes import XPathNode, ElementNode, NamespaceNode, \
    DocumentNode, EtreeElementNode, SchemaElementNode
from elementpath.tree_builders import get_node_tree
from elementpath.xpath_tokens import XPathToken, ValueToken, XPathFunction
from elementpath.serialization import get_serialization_params, serialize_to_xml, \
    serialize_to_json
from elementpath.xpath_context import XPathContext, XPathSchemaContext
from elementpath.regex import translate_pattern, RegexError

from ._xpath30_operators import XPath30Parser
from .xpath30_helpers import UNICODE_DIGIT_PATTERN, DECIMAL_DIGIT_PATTERN, \
    MODIFIER_PATTERN, decimal_to_string, int_to_roman, int_to_alphabetic, \
    format_digits, int_to_words, parse_datetime_picture, parse_datetime_marker, \
    ordinal_suffix

FORMAT_INTEGER_TOKENS = {'A', 'a', 'i', 'I', 'w', 'W', 'Ww'}

DECL_PARAM_PATTERN = re.compile(r'([^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*)\s*=\s*')
EXPONENT_PIC = re.compile(r'\d[eE]\d')

register = XPath30Parser.register
method = XPath30Parser.method
function = XPath30Parser.function


###
# 'inline function' expression or 'function test'

class _InlineFunction(XPathFunction):

    symbol = lookup_name = 'function'
    lbp = 90
    rbp = 90
    label: Union[str, MultiLabel] = MultiLabel('inline function', 'function test')

    body: Optional[XPathToken] = None
    "Body of anonymous inline function."

    variables: Optional[dict[str, Any]] = None
    "In-scope variables linked by let and for expressions and arguments."

    varnames: Optional[list[str]] = None
    "Inline function arguments varnames."

    def __str__(self) -> str:
        return str(self.label)

    @property
    def source(self) -> str:
        if self.label == 'function test':
            if len(self.sequence_types) == 1 and self.sequence_types[0] == '*':
                return 'function(*)'
            else:
                return 'function(%s) as %s' % (
                    ', '.join(self.sequence_types[:-1]), self.sequence_types[-1]
                )

        arguments = []
        return_type = ''
        for var, sq in zip_longest(self, self.sequence_types):
            if var is None:
                if sq != 'item()*':
                    return_type = f' as {sq}'
            elif sq is None or sq == 'item()*':
                arguments.append(var.source)
            else:
                arguments.append(f'{var.source} as {sq}')

        return '%s(%s)%s {%s}' % (
            self.symbol,
            ', '.join(arguments),
            return_type,
            getattr(self.body, 'source', '')
        )

    def __call__(self, *args: ta.FunctionArgType,
                 context: Optional[XPathContext] = None) -> Any:

        def get_argument(v: Any) -> Any:
            if isinstance(v, XPathToken) and not isinstance(v, XPathFunction):
                v = v.evaluate(context)

            if isinstance(v, XPathFunction) and sequence_type.startswith('function('):
                if not v.match_function_test(sequence_type, as_argument=True):
                    msg = "argument {!r}: {} does not match sequence type {}"
                    raise self.error('XPTY0004', msg.format(varname, v, sequence_type))

            elif not match_sequence_type(v, sequence_type, self.parser):
                _v = self.cast_to_primitive_type(v, sequence_type)
                if not match_sequence_type(_v, sequence_type, self.parser):
                    msg = "argument '${}': {} does not match sequence type {}"
                    raise self.error('XPTY0004', msg.format(varname, v, sequence_type))
                return _v
            return v

        sequence_type: str
        self.check_arguments_number(len(args))

        context = copy(context)
        if self.variables and context is not None:
            context.variables.update(self.variables)

        if self.varnames is None:
            self.varnames = []

        assert self.body is not None
        if self.label == 'inline partial function':
            k = 0
            for varname, sequence_type, tk in zip(self.varnames, self.sequence_types, self):
                if context is None:
                    raise self.missing_context()

                if tk.symbol != '?' or tk:
                    context.variables[varname] = tk.evaluate(context)
                else:
                    context.variables[varname] = get_argument(args[k])
                    k += 1

            result = self.body.evaluate(context)
        else:
            if context is None:
                raise self.missing_context()
            elif not args and self:
                if isinstance(context.item, DocumentNode):
                    if isinstance(context.root, DocumentNode):
                        context.item = context.root.getroot()
                    elif context.root is not None:
                        context.item = context.root

                args = cast(tuple[ta.FunctionArgType], (context.item,))

            partial_function = False
            if self.variables is None:
                self.variables = {}

            for varname, sequence_type, value in zip(self.varnames, self.sequence_types, args):
                if isinstance(value, XPathToken) and value.symbol == '?':
                    partial_function = True
                else:
                    context.variables[varname] = get_argument(value)

            if partial_function:
                self.to_partial_function()
                return self

            result = self.body.evaluate(context)

        return self.validated_result(result)

    def nud(self) -> Union[XPathFunction, XPathToken]:  # type: ignore[override]
        def append_sequence_type(tk: XPathToken) -> None:
            if tk.symbol == '(' and len(tk) == 1:
                tk = tk[0]

            sequence_type = tk.source
            next_symbol = self.parser.next_token.symbol
            if sequence_type != 'empty-sequence()' and next_symbol in ('*', '+', '?'):
                self.parser.advance()
                sequence_type += next_symbol
                tk.occurrence = next_symbol

            if not is_sequence_type(sequence_type, self.parser):
                if 'xs:NMTOKENS' in sequence_type \
                        or 'xs:ENTITIES' in sequence_type \
                        or 'xs:IDREFS' in sequence_type:
                    msg = "a list type cannot be used in a function signature"
                    raise self.error('XPST0051', msg)
                raise self.error('XPST0003', "a sequence type expected")

            assert isinstance(self.sequence_types, list)
            self.sequence_types.append(sequence_type)

        if self.parser.next_token.symbol != '(':
            return self.as_name()
        self.parser.advance('(')
        self.sequence_types = []

        if self.parser.next_token.symbol in ('$', ')'):
            self.label = 'inline function'
            self.varnames = []

            while self.parser.next_token.symbol != ')':
                self.parser.next_token.expected('$')
                variable = self.parser.expression(5)
                varname = variable[0].value
                assert isinstance(varname, str)
                if varname in self.varnames:
                    raise self.error('XQST0039')

                self.append(variable)
                self.varnames.append(varname)

                if self.parser.next_token.symbol == 'as':
                    self.parser.advance('as')
                    token = self.parser.expression(90)
                    append_sequence_type(token)
                else:
                    self.sequence_types.append('item()*')

                self.parser.next_token.expected(')', ',')
                if self.parser.next_token.symbol == ',':
                    self.parser.advance()
                    self.parser.next_token.unexpected(')')

            self.parser.advance(')')

        elif self.parser.next_token.symbol == '*':
            self.label = 'function test'
            self.append(self.parser.advance('*'))
            self.sequence_types.append('*')
            self.parser.advance(')')
            return self
        else:
            self.label = 'function test'
            while True:
                token = self.parser.parse_sequence_type()
                append_sequence_type(token)
                self.append(token)
                if self.parser.next_token.symbol != ',':
                    break
                self.parser.advance(',')
            self.parser.advance(')')

        # Add function return sequence type
        if self.parser.next_token.symbol != 'as':
            self.sequence_types.append('item()*')
        else:
            self.parser.advance('as')
            if self.parser.next_token.label not in ('kind test', 'sequence type', 'function test'):
                self.parser.expected_next('(name)', ':')

            token = self.parser.expression(rbp=90)
            append_sequence_type(token)

        if self.label == 'inline function':
            if self.parser.next_token.symbol != '{' and not self:
                self.label = 'function test'
            else:
                self.parser.advance('{')
                if self.parser.next_token.symbol != '}':
                    self.body = self.parser.expression()
                elif self.parser.version >= '3.1':
                    self.body = ValueToken(self.parser, value=[])
                else:
                    raise self.wrong_syntax("inline function has an empty body")
                self.parser.advance('}')

        return self

    def evaluate(self, context: ta.ContextType = None) -> ta.ValueType:
        if context is None:
            raise self.missing_context()
        elif self.label.endswith('function'):
            self.variables = context.variables.copy()  # like a closure
            return self

        # A function test
        if not isinstance(context.item, XPathFunction):
            return []
        elif self.source == 'function(*)':
            return context.item
        elif context.item.match_function_test(self.sequence_types):
            return context.item
        else:
            return []

    def to_partial_function(self) -> None:
        assert self.label != 'function test', "an effective inline function required"

        nargs = len([tk and not tk for tk in self._items if tk.symbol == '?'])
        assert nargs, "a partial function requires at least a placeholder token"

        self._name = None  # noqa
        self.label = 'inline partial function'
        self.nargs = nargs


XPath30Parser.symbol_table['function'] = _InlineFunction


###
# Mathematical functions
@method(function('pi', prefix='math', nargs=0, sequence_types=('xs:double',)))
def evaluate__pi(self: XPathFunction, context: ta.ContextType = None) -> float:
    return math.pi


@method(function('exp', prefix='math', nargs=1, sequence_types=('xs:double?', 'xs:double?')))
def evaluate__exp(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    return math.exp(arg)


@method(function('exp10', prefix='math', nargs=1, sequence_types=('xs:double?', 'xs:double?')))
def evaluate__exp10(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    return float(10 ** arg)


@method(function('log', prefix='math', nargs=1, sequence_types=('xs:double?', 'xs:double?')))
def evaluate__log(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    return float('-inf') if not arg else math.nan if arg <= -1 else math.log(arg)


@method(function('log10', prefix='math', nargs=1, sequence_types=('xs:double?', 'xs:double?')))
def evaluate__log10(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    return float('-inf') if not arg else math.nan if arg <= -1 else math.log10(arg)


@method(function('pow', prefix='math', nargs=2,
                 sequence_types=('xs:double?', 'xs:numeric', 'xs:double?')))
def evaluate__pow(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    if self.context is not None:
        context = self.context

    x = self.get_argument(context, cls=NumericProxy)
    y = self.get_argument(context, index=1, required=True, cls=NumericProxy)
    if x is None:
        return []
    elif not x and y < 0:
        return math.copysign(float('inf'), x) if (y % 2) == 1 else float('inf')

    try:
        return float(x ** y)
    except TypeError:
        return math.nan


@method(function('sqrt', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__sqrt(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    elif arg < 0:
        return math.nan
    return math.sqrt(arg)


@method(function('sin', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__sin(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    elif math.isinf(arg):
        return math.nan
    return math.sin(arg)


@method(function('cos', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__cos(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    elif math.isinf(arg):
        return math.nan
    return math.cos(arg)


@method(function('tan', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__tan(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    elif math.isinf(arg):
        return math.nan
    return math.tan(arg)


@method(function('asin', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__asin(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    elif arg < -1 or arg > 1:
        return math.nan
    return math.asin(arg)


@method(function('acos', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__acos(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    elif arg < -1 or arg > 1:
        return math.nan
    return math.acos(arg)


@method(function('atan', prefix='math', nargs=1,
                 sequence_types=('xs:double?', 'xs:double?')))
def evaluate__atan(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, cls=NumericProxy)
    if arg is None:
        return []
    return math.atan(arg)


@method(function('atan2', prefix='math', nargs=2,
                 sequence_types=('xs:double', 'xs:double', 'xs:double')))
def evaluate__atan2(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[float]:
    if self.context is not None:
        context = self.context

    x = self.get_argument(context, cls=NumericProxy)
    y = self.get_argument(context, index=1, required=True, cls=NumericProxy)
    return math.atan2(x, y)


###
# Formatting functions
@method(function('format-integer', nargs=(2, 3),
                 sequence_types=('xs:integer?', 'xs:string', 'xs:string?', 'xs:string')))
def evaluate__format_integer(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    value = self.get_argument(context, cls=NumericProxy)
    picture = self.get_argument(context, index=1, required=True, cls=str)
    lang = self.get_argument(context, index=2, cls=str)
    if value is None:
        return ''

    if ';' not in picture:
        fmt_token, fmt_modifier = picture, ''
    else:
        fmt_token, fmt_modifier = picture.rsplit(';', 1)

    if MODIFIER_PATTERN.match(fmt_modifier) is None:
        raise self.error('FODF1310')

    if not fmt_token:
        raise self.error('FODF1310')
    elif fmt_token in FORMAT_INTEGER_TOKENS:
        if fmt_token == 'a':
            result = int_to_alphabetic(value, lang)
        elif fmt_token == 'A':
            result = int_to_alphabetic(value, lang).upper()
        elif fmt_token == 'i':
            result = int_to_roman(value).lower()
        elif fmt_token == 'I':
            result = int_to_roman(value)
        elif fmt_token == 'w':
            return int_to_words(value, lang, fmt_modifier)
        elif fmt_token == 'W':
            return int_to_words(value, lang, fmt_modifier).upper()
        else:
            return int_to_words(value, lang, fmt_modifier).title()

    else:
        if UNICODE_DIGIT_PATTERN.search(fmt_token) is None:
            if any(not x.isalpha() and not x.isdigit() for x in fmt_token):
                result = str(value)  # fallback for invalid pictures
            else:
                base_char = '1'
                for base_char in fmt_token:
                    if base_char.isalpha():
                        break
                if base_char.islower():
                    result = int_to_alphabetic(value, base_char)
                else:
                    result = int_to_alphabetic(value, base_char.lower()).upper()

        elif DECIMAL_DIGIT_PATTERN.search(fmt_token) is None or ',,' in fmt_token:
            msg = 'picture argument has an invalid primary format token'
            raise self.error('FODF1310', msg)
        else:
            digits = UNICODE_DIGIT_PATTERN.findall(fmt_token)
            cp = ord(digits[0])
            if any((ord(ch) - cp) > 10 for ch in digits[1:]):
                msg = "picture argument mixes digits from different digit families"
                raise self.error('FODF1310', msg)
            elif fmt_token[0].isdigit():
                if '#' in fmt_token:
                    msg = 'picture argument has an invalid primary format token'
                    raise self.error('FODF1310', msg)
            elif fmt_token[0] != '#':
                raise self.error('FODF1310', "invalid grouping in picture argument")

            if digits[0].isdigit():
                cp = ord(digits[0])
                while chr(cp - 1).isdigit():
                    cp -= 1
                digits_family = ''.join(chr(cp + k) for k in range(10))
            else:
                raise ValueError()

            if value < 0:
                result = '-' + format_digits(str(abs(value)), fmt_token, digits_family)
            else:
                result = format_digits(str(abs(value)), fmt_token, digits_family)

    if fmt_modifier.startswith('o'):
        return f'{result}{ordinal_suffix(value)}'
    return result


@method(function('format-number', nargs=(2, 3),
                 sequence_types=('xs:numeric?', 'xs:string', 'xs:string?', 'xs:string')))
def evaluate__format_number(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    value = self.get_argument(context, cls=NumericProxy)
    picture = self.get_argument(context, index=1, required=True, cls=str)
    decimal_format_name = self.get_argument(context, index=2, cls=str)

    # Check and adapt decimal format name
    if decimal_format_name is not None:
        decimal_format_name = decimal_format_name.strip()
        if decimal_format_name.startswith('Q{'):
            if decimal_format_name.startswith('Q{}'):
                decimal_format_name = decimal_format_name[3:]
            else:
                decimal_format_name = decimal_format_name[1:]
        elif ':' in decimal_format_name:
            try:
                decimal_format_name = get_expanded_name(
                    name=decimal_format_name,
                    namespaces=self.parser.namespaces
                )
            except (KeyError, ValueError):
                raise self.error('FODF1280') from None

    try:
        decimal_format = self.parser.decimal_formats[decimal_format_name]
    except KeyError:
        raise self.error('FODF1280') from None

    pattern_separator = decimal_format['pattern-separator']
    sub_pictures = picture.split(pattern_separator)
    if len(sub_pictures) > 2:
        raise self.error('FODF1310')

    decimal_separator = decimal_format['decimal-separator']
    if any(p.count(decimal_separator) > 1 for p in sub_pictures):
        raise self.error('FODF1310')

    percent_sign = decimal_format['percent']
    per_mille_sign = decimal_format['per-mille']
    if any(p.count(percent_sign) + p.count(per_mille_sign) > 1 for p in sub_pictures):
        raise self.error('FODF1310')

    zero_digit = decimal_format['zero-digit']
    optional_digit = decimal_format['digit']
    digits_family = ''.join(chr(cp + ord(zero_digit)) for cp in range(10))
    if any(optional_digit not in p and all(x not in p for x in digits_family)
           for p in sub_pictures):
        raise self.error('FODF1310')

    grouping_separator = decimal_format['grouping-separator']
    adjacent_pattern = re.compile(r'[\\%s\\%s]{2}' % (grouping_separator, decimal_separator))
    if any(adjacent_pattern.search(p) for p in sub_pictures):
        raise self.error('FODF1310')

    if any(x.endswith(grouping_separator)
           for s in sub_pictures for x in s.split(decimal_separator)):
        raise self.error('FODF1310')

    active_characters = digits_family + ''.join([
        decimal_separator, grouping_separator, pattern_separator, optional_digit
    ])

    exponent_pattern = None
    exponent_separator = 'e'
    if self.parser.version > '3.0':
        # Check optional exponent spec correctness in each sub-picture
        exponent_separator = decimal_format['exponent-separator']
        _pattern = re.compile(r'(?<=[{0}]){1}[{0}]'.format(
            re.escape(active_characters), exponent_separator
        ))
        for p in sub_pictures:
            for match in _pattern.finditer(p):
                if percent_sign in p or per_mille_sign in p:
                    raise self.error('FODF1310')
                elif any(c not in digits_family for c in p[match.span()[1]-1:]):
                    # detailed check to consider suffix
                    has_suffix = False
                    for ch in p[match.span()[1]-1:]:
                        if ch in digits_family:
                            if has_suffix:
                                raise self.error('FODF1310')
                        elif ch in active_characters:
                            raise self.error('FODF1310')
                        else:
                            has_suffix = True

                exponent_pattern = _pattern

    if exponent_pattern is None:
        if any(EXPONENT_PIC.search(s) for s in sub_pictures):
            raise self.error('FODF1310')

    if value is None or math.isnan(value):
        return f"{decimal_format['NaN']}"
    elif isinstance(value, float):
        value = decimal.Decimal.from_float(value)
    elif not isinstance(value, decimal.Decimal):
        value = decimal.Decimal(value)

    minus_sign = decimal_format['minus-sign']

    prefix: str = ''
    if value >= 0:
        subpic = sub_pictures[0]
    else:
        subpic = sub_pictures[-1]
        if len(sub_pictures) == 1:
            prefix = minus_sign

    for k, ch in enumerate(subpic):
        if ch in active_characters:
            prefix += subpic[:k]
            subpic = subpic[k:]
            break
    else:
        prefix += subpic
        subpic = ''

    if not subpic:
        suffix = ''
    elif subpic[-1] == percent_sign:
        suffix = percent_sign
        subpic = subpic[:-1]

        exponent = value.as_tuple().exponent
        if isinstance(exponent, int) and exponent < 0:
            value *= 100
        else:
            value = decimal.Decimal(int(value) * 100)

    elif subpic[-1] == per_mille_sign:
        suffix = per_mille_sign
        subpic = subpic[:-1]

        exponent = value.as_tuple().exponent
        if isinstance(exponent, int) and exponent < 0:
            value *= 1000
        else:
            value = decimal.Decimal(int(value) * 1000)

    else:
        for k, ch in enumerate(reversed(subpic)):
            if ch in active_characters:
                idx = len(subpic) - k
                suffix = subpic[idx:]
                subpic = subpic[:idx]
                break
        else:
            suffix = subpic
            subpic = ''

    exp_fmt = None
    if exponent_pattern is not None:
        exp_match = exponent_pattern.search(subpic)
        if exp_match is not None:
            exp_fmt = subpic[exp_match.span()[0]+1:]
            subpic = subpic[:exp_match.span()[0]]

    fmt_tokens = subpic.split(decimal_separator)
    if all(not fmt for fmt in fmt_tokens):
        raise self.error('FODF1310')

    if math.isinf(value):
        return f"{prefix}{decimal_format['infinity']}{suffix}"

    # Calculate the exponent value if it's in the sub-picture
    exp_value = 0
    if exp_fmt and value:
        num_digits = 0
        for ch in fmt_tokens[0]:
            if ch in digits_family:
                num_digits += 1

        if abs(value) > 1:
            v = abs(value)
            while v > 10 ** num_digits:
                exp_value += 1
                v /= 10

            # modify empty fractional part to store a digit
            if not num_digits:
                if len(fmt_tokens) == 1:
                    fmt_tokens.append(zero_digit)
                elif not fmt_tokens[-1]:
                    fmt_tokens[-1] = zero_digit

        elif len(fmt_tokens) > 1 and fmt_tokens[-1] and value >= 0:
            v = abs(value) * 10
            while v < 10 ** num_digits:
                exp_value -= 1
                v *= 10
        else:
            v = abs(value) * 10
            while v < 10:
                exp_value -= 1
                v *= 10

        if exp_value:
            value = value * decimal.Decimal(10) ** -exp_value

    # round the value by fractional part
    if len(fmt_tokens) == 1 or not fmt_tokens[-1]:
        exp = decimal.Decimal('1')
    else:
        k = -1
        for ch in fmt_tokens[-1]:
            if ch in digits_family or ch == optional_digit:
                k += 1
        exp = decimal.Decimal('.' + '0' * k + '1')

    try:
        if value > 0:
            value = value.quantize(exp, rounding='ROUND_HALF_UP')
        else:
            value = value.quantize(exp, rounding='ROUND_HALF_DOWN')
    except decimal.InvalidOperation:
        pass  # number too large, don't round elementpath..

    chunks = decimal_to_string(value).lstrip('-').split('.')
    kwargs = {
        'digits_family': digits_family,
        'optional_digit': optional_digit,
        'grouping_separator': grouping_separator,
    }
    result = format_digits(chunks[0], fmt_tokens[0], **kwargs)

    if len(fmt_tokens) > 1 and fmt_tokens[-1]:
        has_optional_digit = False
        for ch in fmt_tokens[-1]:
            if ch == optional_digit:
                has_optional_digit = True
            elif ch.isdigit() and has_optional_digit:
                raise self.error('FODF1310')

        if len(chunks) == 1:
            chunks.append(zero_digit)

        decimal_part = format_digits(chunks[1], fmt_tokens[-1], **kwargs)

        for ch in reversed(fmt_tokens[-1]):
            if ch == optional_digit:
                if decimal_part and decimal_part[-1] == zero_digit:
                    decimal_part = decimal_part[:-1]
            else:
                if not decimal_part:
                    decimal_part = zero_digit
                break

        if decimal_part:
            result += decimal_separator + decimal_part

            if not fmt_tokens[0] and result.startswith(zero_digit):
                result = result.lstrip(zero_digit)

    if exp_fmt:
        exp_digits = format_digits(str(abs(exp_value)), exp_fmt, **kwargs)
        if exp_value >= 0:
            result += f'{exponent_separator}{exp_digits}'
        else:
            result += f'{exponent_separator}-{exp_digits}'

    return prefix + result + suffix


function('format-dateTime', nargs=(2, 5),
         sequence_types=('xs:dateTime?', 'xs:string', 'xs:string?',
                         'xs:string?', 'xs:string?', 'xs:string?'))
function('format-date', nargs=(2, 5),
         sequence_types=('xs:date?', 'xs:string', 'xs:string?',
                         'xs:string?', 'xs:string?', 'xs:string?'))
function('format-time', nargs=(2, 5),
         sequence_types=('xs:time?', 'xs:string', 'xs:string?',
                         'xs:string?', 'xs:string?', 'xs:string?'))


@method('format-dateTime')
@method('format-date')
@method('format-time')
def evaluate__format_date_time(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[str]:
    cls: type[Union[DateTime, Date, Time]]
    if self.symbol == 'format-dateTime':
        cls = DateTime
        invalid_markers = ''
    elif self.symbol == 'format-date':
        cls = Date
        invalid_markers = 'HhPmsf'
    else:
        cls = Time
        invalid_markers = 'YMDdFWwCE'

    if self.context is not None:
        context = self.context

    value = self.get_argument(context, cls=cls)
    picture = self.get_argument(context, index=1, required=True, cls=str)
    if len(self) not in [2, 5]:
        raise self.error('XPST0017')
    language = self.get_argument(context, index=2, cls=str)
    calendar = self.get_argument(context, index=3, cls=str)
    place = self.get_argument(context, index=4, cls=str)

    if value is None:
        return ''

    try:
        literals, markers = parse_datetime_picture(picture)
    except ElementPathError as err:
        err.token = self
        raise

    if invalid_markers:
        for mrk in markers:
            if mrk[1] in invalid_markers:
                msg = 'Invalid date formatting component {!r}'.format(mrk)
                raise self.error('FOFD1350', msg)

    result = []
    if language not in ('en', 'it', None):
        language = 'en'
        result.append('[Language: en')

    if calendar is not None:
        if calendar.startswith('Q{}'):
            calendar = calendar[3:]

        if calendar not in ('AD', 'ISO', 'OS'):
            if context is None or calendar != context.default_calendar:
                if QName.is_valid(calendar):
                    if ':' not in calendar:
                        msg = f'unknown calendar in no namespace {calendar!r}'
                        raise self.error('FOFD1340', msg)

                    try:
                        _ = get_expanded_name(calendar, self.parser.namespaces)
                    except (KeyError, ValueError) as err:
                        raise self.error('FOFD1340', str(err)) from None

                elif Patterns.extended_qname.search(calendar) is None:
                    raise self.error('FOFD1340', f'Invalid calendar argument {calendar!r}')
            else:
                result.append('[' if not result else ', ')
                result.append('Calendar: AD')

    if place is not None and zoneinfo is not None:
        try:
            zone = zoneinfo.ZoneInfo(place.strip())
        except zoneinfo.ZoneInfoNotFoundError:
            if not isinstance(context, XPathSchemaContext):
                raise self.error('FOFD1340', f'Invalid place argument {place!r}')
        else:
            value = value.astimezone(zone)

    if result:
        result.append(']')

    for k in range(len(markers)):
        result.append(literals[k])
        try:
            result.append(parse_datetime_marker(markers[k], value, language))
        except ElementPathError as err:
            err.token = self
            raise

    result.append(literals[-1])
    return ''.join(result)


###
# String functions that use regular expressions
@method(function('analyze-string', nargs=(2, 3),
                 sequence_types=('xs:string?', 'xs:string', 'xs:string',
                                 'element(fn:analyze-string-result)')))
def evaluate__analyze_string(self: XPathFunction, context: ta.ContextType = None) \
        -> ElementNode:
    if self.context is not None:
        context = self.context

    input_string = self.get_argument(context, default='', cls=str)
    pattern = self.get_argument(context, 1, required=True, cls=str)
    flags = 0
    if len(self) > 2:
        for c in self.get_argument(context, 2, required=True, cls=str):
            if c in 'smix':
                flags |= getattr(re, c.upper())
            elif c == 'q' and self.parser.version > '2':
                pattern = re.escape(pattern)
            else:
                raise self.error('FORX0001', "Invalid regular expression flag %r" % c)

    try:
        python_pattern = translate_pattern(pattern, flags, self.parser.xsd_version)
        compiled_pattern = re.compile(python_pattern, flags=flags)
    except (re.error, RegexError) as err:
        msg = "Invalid regular expression: {}"
        raise self.error('FORX0002', msg.format(str(err))) from None
    except OverflowError as err:
        raise self.error('FORX0002', err) from None

    if compiled_pattern.match('') is not None:
        raise self.error('FORX0003', "pattern matches a zero-length string")

    if context is None:
        raise self.missing_context()

    level = 0
    escaped = False
    char_class = False
    group_levels = [0]
    for s in compiled_pattern.pattern:
        if escaped:
            escaped = False
        elif s == '\\':
            escaped = True
        elif char_class:
            if s == ']':
                char_class = False
        elif s == '[':
            char_class = True
        elif s == '(':
            group_levels.append(level)
            level += 1
        elif s == ')':
            level -= 1

    lines = ['<analyze-string-result xmlns="{}">'.format(XPATH_FUNCTIONS_NAMESPACE)]
    k = 0

    while k < len(input_string):
        match = compiled_pattern.search(input_string, k)
        if match is None:
            lines.append('<non-match>{}</non-match>'.format(input_string[k:]))
            break
        elif not match.groups():
            start, stop = match.span()
            if start > k:
                lines.append('<non-match>{}</non-match>'.format(input_string[k:start]))
            lines.append('<match>{}</match>'.format(input_string[start:stop]))
            k = stop
        else:
            start, stop = match.span()
            if start > k:
                lines.append('<non-match>{}</non-match>'.format(input_string[k:start]))
                k = start

            match_items = []
            group_tmpl = '<group nr="{}">{}'
            empty_group_tmpl = '<group nr="{}"/>'
            unclosed_groups = 0

            for idx in range(1, compiled_pattern.groups + 1):
                _start, _stop = match.span(idx)
                if _start < 0:
                    continue
                elif _start > k:
                    if unclosed_groups:
                        for _ in range(unclosed_groups):
                            match_items.append('</group>')
                        unclosed_groups = 0

                    match_items.append(input_string[k:_start])

                if _start == _stop:
                    if group_levels[idx] <= group_levels[idx - 1]:
                        for _ in range(unclosed_groups):
                            match_items.append('</group>')
                        unclosed_groups = 0
                    match_items.append(empty_group_tmpl.format(idx))
                    k = _stop
                elif idx == compiled_pattern.groups:
                    k = _stop
                    match_items.append(group_tmpl.format(idx, input_string[_start:k]))
                    match_items.append('</group>')
                else:
                    next_start = match.span(idx + 1)[0]
                    if next_start < 0 or _stop < next_start or _stop == next_start \
                            and group_levels[idx + 1] <= group_levels[idx]:
                        k = _stop
                        match_items.append(group_tmpl.format(idx, input_string[_start:k]))
                        match_items.append('</group>')
                    else:
                        k = next_start
                        match_items.append(group_tmpl.format(idx, input_string[_start:k]))
                        unclosed_groups += 1

            for _ in range(unclosed_groups):
                match_items.append('</group>')

            match_items.append(input_string[k:stop])
            k = stop
            lines.append('<match>{}</match>'.format(''.join(match_items)))

    lines.append('</analyze-string-result>')
    if self.parser.defuse_xml:
        root = context.etree.XML(defuse_xml(''.join(lines)))
    else:
        root = context.etree.XML(''.join(lines))

    return cast(ElementNode, get_node_tree(root=root, namespaces=self.parser.namespaces))


###
# Functions and operators on nodes
@method(function('path', nargs=(0, 1), sequence_types=('node()?', 'xs:string?')))
def evaluate__path(self: XPathFunction, context: ta.ContextType = None) -> ta.OneOrEmpty[str]:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    if not self:
        item = context.item
    else:
        item = self.get_argument(context)
        if item is None:
            return []

    if not isinstance(item, XPathNode):
        return []
    elif context.root is None or (root_node := item.root_node) is not context.root:
        # The context has no root or the root is not the root of the item node
        return []
    elif not isinstance(root_node, (DocumentNode, SchemaElementNode)):
        # It's a fragment: add fn:root() to select the root position
        path = item.path[len(root_node.path):]
        return f"Q{{{XPATH_FUNCTIONS_NAMESPACE}}}root(){path}"
    else:
        return item.path or []


@method(function('has-children', nargs=(0, 1), sequence_types=('node()?', 'xs:boolean')))
def evaluate__has_children(self: XPathFunction, context: ta.ContextType = None) -> bool:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    if not self:
        item = context.item
        if not isinstance(item, XPathNode):
            raise self.error('XPTY0004', 'context item must be a node')
    else:
        item = self.get_argument(context)
        if item is None:
            return False
        elif not isinstance(item, XPathNode):
            raise self.error('XPTY0004', 'argument must be a node')

    return isinstance(item, DocumentNode) or \
        isinstance(item, EtreeElementNode) and (len(item.value) > 0 or item.value.text is not None)


@method(function('innermost', nargs=1, sequence_types=('node()*', 'node()*')))
def select__innermost(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[XPathNode]:
    if context is None:
        raise self.missing_context()

    nodes = [e for e in self[0].select(context)]
    if any(not isinstance(x, XPathNode) for x in nodes):
        raise self.error('XPTY0004', 'argument must contain only nodes')

    ancestors = {x for context.item in nodes for x in context.iter_ancestors(axis='ancestor')}
    results = {x for x in nodes if x not in ancestors}
    yield from cast(list[XPathNode], sorted(results, key=node_position))


@method(function('outermost', nargs=1, sequence_types=('node()*', 'node()*')))
def select__outermost(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[XPathNode]:
    if context is None:
        raise self.missing_context()

    nodes: Union[list[ta.ItemType], set[ta.ItemType]]
    nodes = [e for e in self[0].select(context)]
    if any(not isinstance(x, XPathNode) for x in nodes):
        raise self.error('XPTY0004', 'argument must contain only nodes')

    results = set()
    if len(nodes) > 10:
        nodes = set(nodes)

    for item in nodes:
        context.item = item
        ancestors = {x for x in context.iter_ancestors(axis='ancestor')}
        if any(x in nodes for x in ancestors):
            continue
        results.add(item)

    yield from cast(list[XPathNode], sorted(results, key=node_position))


##
# Functions and operators on sequences
@method(function('head', nargs=1, sequence_types=('item()*', 'item()?')))
def evaluate__head(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[ta.ItemType]:
    for item in self[0].select(self.context or context):
        return item
    else:
        return []


@method(function('tail', nargs=1, sequence_types=('item()*', 'item()*')))
def select__tail(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    for k, item in enumerate(self[0].select(self.context or context)):
        if k:
            yield item


@method(function('generate-id', nargs=(0, 1), sequence_types=('node()?', 'xs:string')))
def evaluate__generate_id(self: XPathFunction, context: ta.ContextType = None) -> str:
    arg: ta.NumericType | None
    arg = self.get_argument(self.context or context, default_to_context=True)
    if arg is None:
        return ''
    elif not isinstance(arg, XPathNode):
        if self:
            raise self.error('XPTY0004', "argument is not a node")
        raise self.error('XPTY0004', "context item is not a node")
    else:
        return f'ID{id(arg)}'


@method(function('uri-collection', nargs=(0, 1),
                 sequence_types=('xs:string?', 'xs:anyURI*')))
def evaluate__uri_collection(self: XPathFunction, context: ta.ContextType = None) \
        -> list[AnyURI] | list[NoReturn]:
    if self.context is not None:
        context = self.context

    uri = self.get_argument(context)
    if context is None:
        raise self.missing_context()
    elif isinstance(context, XPathSchemaContext):
        return []
    elif not self or uri is None:
        if context.default_resource_collection is None:
            raise self.error('FODC0002', 'no default resource collection has been defined')
        resource_collection = [AnyURI(context.default_resource_collection)]
    else:
        try:
            AnyURI(uri)
        except ValueError:
            raise self.error('FODC0004', 'invalid argument to fn:uri-collection') from None

        if not context.resource_collections:
            resource_collection = []
        else:
            uri = self.get_absolute_uri(uri)
            try:
                resource_collection = [AnyURI(x) for x in context.resource_collections[uri]]
            except (KeyError, TypeError):
                url_parts = urlsplit(uri)
                if url_parts.scheme in ('', 'file') and \
                        not url_parts.path.startswith(':') and url_parts.path.endswith('/'):
                    raise self.error('FODC0003', 'collection URI is a directory')
                raise self.error('FODC0002', '{!r} collection not found'.format(uri)) from None

    if not match_sequence_type(resource_collection, 'xs:anyURI*', self.parser):
        raise self.error('XPDY0050', "type does not match sequence type xs:anyURI*")

    return resource_collection


@method(function('unparsed-text', nargs=(1, 2),
                 sequence_types=('xs:string?', 'xs:string', 'xs:string?')))
@method(function('unparsed-text-lines', nargs=(1, 2),
                 sequence_types=('xs:string?', 'xs:string', 'xs:string*')))
def evaluate__unparsed_text(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[str]:
    if self.context is not None:
        context = self.context

    href: Optional[str] = self.get_argument(context, cls=str)
    if href is None:
        return []
    elif urlsplit(href).fragment:
        raise self.error('FOUT1170')

    encoding: str
    if len(self) > 1:
        encoding = self.get_argument(context, index=1, required=True, cls=str)
    else:
        encoding = 'UTF-8'

    try:
        uri = self.get_absolute_uri(href)
    except ValueError:
        raise self.error('FOUT1170') from None

    try:
        codecs.lookup(encoding)
    except LookupError:
        raise self.error('FOUT1190') from None

    if context is not None and uri in context.text_resources:
        text = context.text_resources[uri]
    else:
        try:
            with urlopen(uri) as rp:
                stream_reader = codecs.getreader(encoding)(rp)
                text = stream_reader.read()
        except URLError as err:
            raise self.error('FOUT1170', err) from None
        except ValueError as err:
            if len(self) > 1:
                raise self.error('FOUT1190', err) from None

            try:
                with urlopen(uri) as rp:
                    stream_reader = codecs.getreader('UTF-16')(rp)
                    text = stream_reader.read()
            except URLError as err:
                raise self.error('FOUT1170', err) from None
            except ValueError as err:
                raise self.error('FOUT1190', err) from None

        if context is not None:
            context.text_resources[uri] = text

    if not all(is_xml_codepoint(ord(s)) for s in text):
        raise self.error('FOUT1190')

    text = text.lstrip('\ufeff')

    if self.symbol == 'unparsed-text-lines':
        lines = Patterns.xml_newlines.split(text)
        return lines[:-1] if lines[-1] == '' else lines

    return text


@method(function('unparsed-text-available', nargs=(1, 2),
                 sequence_types=('xs:string?', 'xs:string', 'xs:boolean')))
def evaluate__unparsed_text_available(self: XPathFunction, context: ta.ContextType = None) \
        -> bool:
    if self.context is not None:
        context = self.context

    href = self.get_argument(context, cls=str)
    if href is None:
        return False
    elif urlsplit(href).fragment:
        return False

    if len(self) > 1:
        encoding = self.get_argument(context, index=1, required=True, cls=str)
    else:
        encoding = 'UTF-8'

    try:
        uri = self.get_absolute_uri(href)
    except ValueError:
        return False

    try:
        codecs.lookup(encoding)
    except LookupError:
        return False

    try:
        with urlopen(uri) as rp:
            stream_reader = codecs.getreader(encoding)(rp)
            for line in stream_reader:
                if any(not is_xml_codepoint(ord(s)) for s in line):
                    return False
    except URLError:
        return False
    except ValueError:
        if len(self) > 1:
            return False
    else:
        return True

    # Fallback auto-detection with utf-16
    try:
        with urlopen(uri) as rp:
            stream_reader = codecs.getreader('UTF-16')(rp)
            for line in stream_reader:
                if any(not is_xml_codepoint(ord(s)) for s in line):
                    return False
    except (ValueError, URLError):
        return False
    else:
        return True


@method(function('environment-variable', nargs=1,
                 sequence_types=('xs:string', 'xs:string?')))
def evaluate__environment_variable(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[str]:
    if self.context is not None:
        context = self.context

    name: str = self.get_argument(context, required=True, cls=str)
    if context is None:
        raise self.missing_context()
    elif not context.allow_environment:
        return []
    else:
        value = os.environ.get(name)
        return value if value is not None else []


@method(function('available-environment-variables', nargs=0,
                 sequence_types=('xs:string*',)))
def evaluate__available_env_vars(self: XPathFunction, context: ta.ContextType = None) \
        -> list[str] | list[NoReturn]:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    if not context.allow_environment:
        return []
    else:
        return xlist(os.environ)


###
# Parsing and serializing
@method(function('parse-xml', nargs=1,
                 sequence_types=('xs:string?', 'document-node(element(*))?')))
def evaluate__parse_xml(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[DocumentNode]:
    # TODO: resolve relative entity references with static base URI
    if self.context is not None:
        context = self.context

    arg: Optional[str] = self.get_argument(context, cls=str)
    if arg is None:
        return []
    elif context is None:
        raise self.missing_context()

    etree = context.etree
    try:
        if self.parser.defuse_xml:
            root = etree.XML(defuse_xml(arg.encode('utf-8')))
        else:
            root = etree.XML(arg.encode('utf-8'))
    except etree.ParseError:
        raise self.error('FODC0006')
    else:
        return cast(DocumentNode, get_node_tree(etree.ElementTree(root), self.parser.namespaces))


@method(function('parse-xml-fragment', nargs=1,
                 sequence_types=('xs:string?', 'document-node()?')))
def evaluate__parse_xml_fragment(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[DocumentNode]:
    if self.context is not None:
        context = self.context

    arg: Optional[str] = self.get_argument(context, cls=str)
    if arg is None or isinstance(context, XPathSchemaContext):
        return []
    elif context is None:
        raise self.missing_context()

    # Wrap argument in a fake document because an
    # XML document can have only one root element
    if arg.startswith('<?xml '):
        xml_declaration, _, arg = arg[6:].partition('?>')
        xml_params = DECL_PARAM_PATTERN.findall(xml_declaration)
        if 'encoding' not in xml_params:
            raise self.error('FODC0006', "'encoding' argument is mandatory")

        for param in xml_params:
            if param not in ('version', 'encoding'):
                msg = f'unexpected parameter {param!r} in XML declaration'
                raise self.error('FODC0006', msg)

    if arg.lstrip().startswith('<!DOCTYPE'):
        raise self.error('FODC0006', "<!DOCTYPE is not allowed")

    etree = context.etree
    try:
        if self.parser.defuse_xml:
            root = etree.XML(defuse_xml(arg))
        else:
            root = etree.XML(arg)
    except etree.ParseError as err:
        # A not parsable fragment: try to parse including XML data in a dummy element.
        try:
            dummy_element_node = get_node_tree(
                root=etree.XML(f'<document>{arg}</document>'),
                namespaces=self.parser.namespaces
            )
        except etree.ParseError:
            raise self.error('FODC0006', str(err)) from None
        else:
            assert isinstance(dummy_element_node, ElementNode)
            return dummy_element_node.get_document_node(replace=True)
    else:
        return cast(DocumentNode, get_node_tree(
            root=etree.ElementTree(root),
            namespaces=self.parser.namespaces
        ))


@method(function('serialize', nargs=(1, 2), sequence_types=(
        'item()*', 'element(output:serialization-parameters)?', 'xs:string')))
def evaluate__serialize(self: XPathFunction, context: ta.ContextType = None) -> str:
    # TODO full implementation of serialization with
    #  https://www.w3.org/TR/xpath-functions-30/#xslt-xquery-serialization-30
    if self.context is not None:
        context = self.context

    params = self.get_argument(context, index=1) if len(self) == 2 else None
    kwargs = get_serialization_params(params, token=self)

    if context is None:
        raise self.missing_context()
    elif isinstance(context, XPathSchemaContext):
        return ''  # not applicable to schemas

    method_ = kwargs.get('method', 'xml')
    if method_ in ('xml', 'html', 'text'):
        etree_module = context.etree
        if context.namespaces:
            for pfx, uri in context.namespaces.items():
                etree_module.register_namespace(pfx, uri)
        else:
            for pfx, uri in self.parser.namespaces.items():
                etree_module.register_namespace(pfx, uri)

        return serialize_to_xml(self[0].select(context), etree_module, **kwargs)

    elif method_ == 'json':
        return serialize_to_json(self[0].select(context), token=self, **kwargs)
    else:
        return ''


###
# Higher-order functions

@method(function('function-lookup', nargs=2,
                 sequence_types=('xs:QName', 'xs:integer', 'function(*)?')))
def evaluate__function_lookup(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[XPathFunction]:
    if self.context is not None:
        context = self.context

    qname: QName = self.get_argument(context, cls=QName, required=True)
    arity: int = self.get_argument(context, index=1, cls=int, required=True)
    if qname.namespace == '':
        return []

    try:
        cls = self.parser.symbol_table[qname.expanded_name]
    except KeyError:
        try:
            cls = self.parser.symbol_table[qname.local_name]
        except KeyError:
            return []

    assert issubclass(cls, XPathFunction)
    try:
        func = cls(self.parser, nargs=arity)
    except TypeError:
        return []

    func.namespace = qname.namespace
    func.context = copy(context)
    return func


@method(function('function-name', nargs=1, sequence_types=('function(*)', 'xs:QName?')))
def evaluate__function_name(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[QName]:
    if self.context is not None:
        context = self.context

    if isinstance(self[0], XPathFunction):
        func = self[0]
    else:
        func = self.get_argument(context)

    if not isinstance(func, XPathFunction):
        raise self.error('XPTY0004', "argument is not a function")
    else:
        name = func.qname
        return [] if name is None else name


@method(function('function-arity', nargs=1, sequence_types=('function(*)', 'xs:integer')))
def evaluate__function_arity(self: XPathFunction, context: ta.ContextType = None) -> int:
    if isinstance(self[0], XPathFunction):
        return self[0].arity

    func: XPathFunction
    func = self.get_argument(self.context or context, cls=XPathFunction, required=True)
    return func.arity


@method(function('for-each', nargs=2,
                 sequence_types=('item()*', 'function(item()) as item()*', 'item()*')))
def select__for_each(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    if self.context is not None:
        context = self.context

    func = self[1][1] if self[1].symbol == ':' else self[1]
    if not isinstance(func, XPathFunction):
        func = self.get_argument(context, index=1, cls=XPathFunction, required=True)
    assert isinstance(func, XPathFunction)

    for item in self[0].select(context):
        result = func(item, context=context)
        if isinstance(result, list):
            yield from result
        else:
            yield result


@method(function('filter', nargs=2,
                 sequence_types=('item()*', 'function(item()) as xs:boolean', 'item()*')))
def select__filter(self: XPathFunction, context: ta.ContextType = None)\
        -> Iterator[ta.ItemType]:
    func = self[1][1] if self[1].symbol == ':' else self[1]
    if not isinstance(func, XPathFunction):
        func = self.get_argument(context, index=1, cls=XPathFunction, required=True)
    assert isinstance(func, XPathFunction)

    if func.nargs == 0:
        raise self.error('XPTY0004', f'invalid number of arguments {func.nargs}')

    for item in self[0].select(context):
        cond = func(item, context=context)
        if not isinstance(cond, bool):
            raise self.error('XPTY0004', 'a single boolean value required')
        if cond:
            yield item


@method(function('fold-left', nargs=3,
                 sequence_types=('item()*', 'item()*',
                                 'function(item()*, item()) as item()*', 'item()*')))
def select__fold_left(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    func = self[2][1] if self[2].symbol == ':' else self[2]
    if not isinstance(func, XPathFunction):
        func = self.get_argument(context, index=2, cls=XPathFunction, required=True)
    assert isinstance(func, XPathFunction)

    if func.arity != 2:
        raise self.error('XPTY0004', "function arity must be 2")

    zero = self.get_argument(context, index=1)

    result = zero
    for item in self[0].select(context):
        result = func(result, item, context=context)

    if isinstance(result, list):
        yield from result
    else:
        yield result


@method(function('fold-right', nargs=3,
                 sequence_types=('item()*', 'item()*',
                                 'function(item()*, item()) as item()*', 'item()*')))
def select__fold_right(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    func = self[2][1] if self[2].symbol == ':' else self[2]
    if not isinstance(func, XPathFunction):
        func = self.get_argument(context, index=2, cls=XPathFunction, required=True)
    assert isinstance(func, XPathFunction)

    if func.arity != 2:
        raise self.error('XPTY0004', "function arity must be 2")

    zero = self.get_argument(context, index=1)

    result = zero
    sequence = [x for x in self[0].select(context)]

    for item in reversed(sequence):
        result = func(item, result, context=context)

    if isinstance(result, list):
        yield from result
    else:
        yield result


@method(function('for-each-pair', nargs=3,
                 sequence_types=('item()*', 'item()*',
                                 'function(item(), item()) as item()*', 'item()*')))
def select__for_each_pair(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    func = self[2][1] if self[2].symbol == ':' else self[2]
    if not isinstance(func, XPathFunction):
        func = self.get_argument(context, index=2, cls=XPathFunction, required=True)

    if not isinstance(func, XPathFunction):
        raise self.error('XPTY0004', "invalid type for 3rd argument {!r}".format(func))
    elif func.arity != 2:
        raise self.error('XPTY0004', "function arity of 3rd argument must be 2")

    for item1, item2 in zip(self[0].select(context), self[1].select(context)):
        result = func(item1, item2, context=context)
        if isinstance(result, list):
            yield from result
        else:
            yield result


@method(function('namespace-node', nargs=0, label='kind test'))
def select__namespace_node_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[NamespaceNode]:
    if context is None:
        raise self.missing_context()
    elif isinstance(context.item, NamespaceNode):
        yield context.item
    elif isinstance(context, XPathSchemaContext):
        return  # deprecated for XP20+ and not needed for schema analysis
    elif isinstance(context.item, ElementNode):
        elem = context.item
        for context.item in elem.namespace_nodes:
            yield context.item  # noqa


###
# Redefined or extended functions
XPath30Parser.unregister('data')
XPath30Parser.unregister('document-uri')
XPath30Parser.unregister('nilled')
XPath30Parser.unregister('node-name')
XPath30Parser.unregister('string-join')
XPath30Parser.unregister('round')


@method(function('data', nargs=(0, 1), sequence_types=('item()*', 'xs:anyAtomicType*')))
def select__data(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ta.AtomicType]:
    if self.context is not None:
        context = self.context

    if self:
        yield from self[0].atomization(context)
    elif context is None:
        raise self.missing_context()
    else:
        yield from self.atomize_item(context.item)


@method(function('document-uri', nargs=(0, 1), sequence_types=('node()?', 'xs:anyURI?')))
def evaluate__document_uri(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[AnyURI]:
    if self.context is not None:
        context = self.context
    elif context is None:
        raise self.missing_context()

    arg: ta.NumericType | None = self.get_argument(context, default_to_context=True)
    if isinstance(arg, DocumentNode):
        uri = arg.document_uri
        if uri is not None:
            return AnyURI(uri)
        elif isinstance(context.root, DocumentNode):
            if context.documents:
                for uri, doc in context.documents.items():
                    if doc and doc.document is context.root.document:
                        return AnyURI(uri)
    return []


@method(function('nilled', nargs=(0, 1), sequence_types=('node()?', 'xs:boolean?')))
def evaluate__nilled(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[bool]:
    arg: ta.NumericType | None = self.get_argument(self.context or context, default_to_context=True)
    if arg is None:
        return []
    elif isinstance(arg, XPathNode):
        result = arg.nilled
        return result if result is not None else []
    else:
        raise self.error('XPTY0004', 'an XPath node required')


@method(function('node-name', nargs=(0, 1), sequence_types=('node()?', 'xs:QName?')))
def evaluate__node_name(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[QName]:
    arg: ta.NumericType | None
    arg = self.get_argument(self.context or context, default_to_context=True)
    if arg is None:
        return []
    elif isinstance(arg, XPathNode):
        name = arg.name
        if name is None:
            return []
        elif name.startswith('{'):
            # name is a QName in extended format
            namespace, local_name = split_expanded_name(name)
            if not namespace:
                return QName('', local_name)

            for pfx, uri in self.parser.namespaces.items():
                if uri == namespace:
                    if not pfx:
                        return QName(uri, local_name)
                    return QName(uri, '{}:{}'.format(pfx, local_name))
            raise self.error('FONS0004', 'no prefix found for namespace {}'.format(namespace))
        else:
            # name is a local name
            return QName(self.parser.namespaces.get('', ''), name)
    else:
        raise self.error('XPTY0004', 'an XPath node required')


@method(function('string-join', nargs=(1, 2),
                 sequence_types=('xs:string*', 'xs:string', 'xs:string')))
def evaluate__string_join(self: XPathFunction, context: ta.ContextType = None) -> str:
    if self.context is not None:
        context = self.context

    items = [
        self.validated_value(s, cls=str, promote=AnyURI, index=k)
        for k, s in enumerate(self[0].atomization(context))
    ]

    if len(self) == 1:
        return ''.join(items)
    separator: str = self.get_argument(context, 1, required=True, cls=str)
    return separator.join(items)


@method(function('round', nargs=(1, 2),
                 sequence_types=('xs:numeric?', 'xs:integer', 'xs:numeric?')))
def evaluate__round(self: XPathFunction, context: ta.ContextType = None) \
        -> ta.OneNumericOrEmpty:
    if self.context is not None:
        context = self.context

    arg: ta.NumericType | None = self.get_argument(context)
    if arg is None:
        return []
    elif isinstance(arg, XPathNode) or self.parser.compatibility_mode:
        arg = self.number_value(arg)

    if isinstance(arg, float) and (math.isnan(arg) or math.isinf(arg)):
        return arg

    precision: int = self.get_argument(context, index=1, default=0, cls=int)
    try:
        if precision < 0:
            return type(arg)(round(arg, precision))  # type: ignore[call-overload, arg-type]

        number = decimal.Decimal(arg)
        exponent = decimal.Decimal('1') / 10 ** precision
        if number > 0:
            return type(arg)(number.quantize(exponent, rounding='ROUND_HALF_UP'))
        else:
            return type(arg)(number.quantize(exponent, rounding='ROUND_HALF_DOWN'))
    except TypeError as err:
        if isinstance(context, XPathSchemaContext):
            return []
        raise self.error('FORG0006', err) from None
    except decimal.InvalidOperation:
        if isinstance(arg, str):
            if isinstance(context, XPathSchemaContext):
                return []
            raise self.error('XPTY0004') from None
        return round(arg)
    except decimal.DecimalException as err:
        if isinstance(context, XPathSchemaContext):
            return []
        raise self.error('FOCA0002', err) from None



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
A unified loader module for ElementTree with a safe parser and helper functions.
"""
import re
import io
import importlib
from collections import Counter
from collections.abc import Iterator, MutableMapping
from pyexpat import XMLParserType
from typing import cast, Any, Optional, Union
from xml.dom import pulldom
from xml.etree import ElementTree
from xml.sax import SAXParseException
from xml.sax import expatreader  # type: ignore[attr-defined, unused-ignore]

from elementpath.protocols import ElementProtocol, DocumentProtocol, EtreeElementProtocol, \
    LxmlElementProtocol
from elementpath.exceptions import XMLResourceForbidden, ElementPathOSError

ElementType = Union[EtreeElementProtocol, LxmlElementProtocol, ElementTree.Element]


class SafeExpatParser(expatreader.ExpatParser):  # type: ignore[misc, unused-ignore]
    _parser: XMLParserType

    def forbid_entity_declaration(self, name, is_parameter_entity,  # type: ignore
                                  value, base, sysid, pubid, notation_name):
        raise XMLResourceForbidden(f"Entities are forbidden (entity_name={name!r})")

    def forbid_unparsed_entity_declaration(self, name, base,  # type: ignore
                                           sysid, pubid, notation_name):
        raise XMLResourceForbidden(f"Unparsed entities are forbidden (entity_name={name!r})")

    def forbid_external_entity_reference(self, context, base, sysid, pubid):  # type: ignore
        raise XMLResourceForbidden(
            f"External references are forbidden (system_id={sysid!r}, public_id={pubid!r})"
        )  # pragma: no cover

    def reset(self) -> None:
        super().reset()
        self._parser.EntityDeclHandler = self.forbid_entity_declaration
        self._parser.UnparsedEntityDeclHandler = self.forbid_unparsed_entity_declaration
        self._parser.ExternalEntityRefHandler = self.forbid_external_entity_reference


def defuse_xml(xml_source: Union[str, bytes]) -> Union[str, bytes]:
    resource: Any
    if isinstance(xml_source, str):
        resource = io.StringIO(xml_source)
    else:
        resource = io.BytesIO(xml_source)

    parser = SafeExpatParser()
    try:
        for event, node in pulldom.parse(resource, parser):
            if event == pulldom.START_ELEMENT:
                break
    except SAXParseException:
        pass  # the purpose is to defuse not to check xml source syntax
    except OSError as err:
        raise ElementPathOSError(str(err))

    return xml_source


def is_etree_element(obj: Any) -> bool:
    return hasattr(obj, 'tag') and hasattr(obj, 'attrib') and hasattr(obj, 'text')


def is_lxml_etree_element(obj: Any) -> bool:
    return is_etree_element(obj) and \
        hasattr(obj, 'getparent') and \
        hasattr(obj, 'nsmap') and \
        obj.__class__.__module__ in ('lxml.etree', 'lxml.html')


def is_etree_element_instance(obj: Any) -> bool:
    """Strictly checks that the objects is an ElementTree or lxml.etree Element."""
    return isinstance(obj, ElementTree.Element) or is_lxml_etree_element(obj)


def is_etree_document(obj: Any) -> bool:
    return hasattr(obj, 'getroot') and hasattr(obj, 'parse') and hasattr(obj, 'iter')


def is_lxml_etree_document(obj: Any) -> bool:
    return is_etree_document(obj) and \
        hasattr(obj, 'xpath') and \
        hasattr(obj, 'xslt') and \
        obj.__class__.__module__ in ('lxml.etree', 'lxml.html')


def is_etree_document_instance(obj: Any) -> bool:
    """Strictly checks that the objects is an ElementTree or lxml.etree document."""
    return isinstance(obj, ElementTree.ElementTree) or is_lxml_etree_document(obj)


def etree_iter_strings(elem: Union[DocumentProtocol, ElementProtocol],
                       normalize: bool = False) -> Iterator[str]:
    e: ElementProtocol
    if normalize:
        if hasattr(elem, 'getroot'):
            root = cast(DocumentProtocol, elem).getroot()
            if root is None:
                return
        else:
            root = elem

        for e in elem.iter():
            if callable(e.tag):
                continue
            if e.text is not None:
                yield e.text.strip() if e is root else e.text
            if e.tail is not None and e is not root:
                yield e.tail.strip() if e in root else e.tail
    else:
        for e in elem.iter():
            if callable(e.tag):
                continue
            if e.text is not None:
                yield e.text
            if e.tail is not None and e is not elem:
                yield e.tail


def etree_deep_equal(e1: ElementProtocol, e2: ElementProtocol) -> bool:
    if e1.tag != e2.tag:
        return False
    elif (e1.text or '').strip() != (e2.text or '').strip():
        return False
    elif (e1.tail or '').strip() != (e2.tail or '').strip():
        return False
    elif e1.attrib != e2.attrib:
        return False
    elif len(e1) != len(e2):
        return False
    return all(etree_deep_equal(c1, c2) for c1, c2 in zip(e1, e2))


def etree_iter_paths(elem: ElementProtocol, path: str = '.') \
        -> Iterator[tuple[ElementProtocol, str]]:

    yield elem, path
    comment_nodes = 0
    pi_nodes = Counter[Optional[str]]()
    positions = Counter[Optional[str]]()

    for child in elem:
        if callable(child.tag):
            if child.tag.__name__ == 'Comment':
                comment_nodes += 1
                yield child, f'{path}/comment()[{comment_nodes}]'
                continue

            try:
                name = cast(str, child.target)  # type: ignore[attr-defined]
            except AttributeError:
                assert child.text is not None
                name = child.text.split(' ', maxsplit=1)[0]

            pi_nodes[name] += 1
            yield child, f'{path}/processing-instruction({name})[{pi_nodes[name]}]'
            continue

        if child.tag.startswith('{'):
            tag = f'Q{child.tag}'
        else:
            tag = f'Q{{}}{child.tag}'

        if path == '/':
            child_path = f'/{tag}'
        elif path:
            child_path = '/'.join((path, tag))
        else:
            child_path = tag

        positions[child.tag] += 1
        child_path += f'[{positions[child.tag]}]'

        yield from etree_iter_paths(child, child_path)


def etree_tostring(elem: ElementType,
                   namespaces: Optional[MutableMapping[str, str]] = None,
                   indent: str = '',
                   max_lines: Optional[int] = None,
                   spaces_for_tab: Optional[int] = 4,
                   xml_declaration: Optional[bool] = None,
                   encoding: str = 'unicode',
                   method: str = 'xml') -> Union[str, bytes]:
    """
    Serialize an Element tree to a string.

    :param elem: the Element instance.
    :param namespaces: is an optional mapping from namespace prefix to URI. \
    Provided namespaces are registered before serialization. Ignored if the \
    provided *elem* argument is a lxml Element instance.
    :param indent: the baseline indentation.
    :param max_lines: if truncate serialization after a number of lines \
    (default: do not truncate).
    :param spaces_for_tab: number of spaces for replacing tab characters. For \
    default tabs are replaced with 4 spaces, provide `None` to keep tab characters.
    :param xml_declaration: if set to `True` inserts the XML declaration at the head.
    :param encoding: if "unicode" (the default) the output is a string, \
    otherwise it’s binary.
    :param method: is either "xml" (the default), "html" or "text".
    :return: a Unicode string.
    """
    def reindent(line: str) -> str:
        if not line:
            return line
        elif line.startswith(min_indent):
            return line[start:] if start >= 0 else indent[start:] + line
        else:
            return indent + line

    etree_module: Any
    if not is_etree_element_instance(elem):
        raise TypeError(f"{elem!r} is not an Element")
    elif not hasattr(elem, 'nsmap'):
        etree_module = ElementTree
    else:
        etree_module = importlib.import_module('lxml.etree')

    if namespaces and not hasattr(elem, 'nsmap'):
        default_namespace = namespaces.get('')
        for prefix, uri in namespaces.items():
            if prefix and not re.match(r'ns\d+$', prefix):
                etree_module.register_namespace(prefix, uri)
                if uri == default_namespace:
                    default_namespace = None

        if default_namespace:
            etree_module.register_namespace('', default_namespace)

    xml_text = etree_module.tostring(elem, encoding=encoding, method=method)
    if isinstance(xml_text, bytes):
        xml_text = xml_text.decode('utf-8')

    if spaces_for_tab is not None:
        xml_text = xml_text.replace('\t', ' ' * spaces_for_tab)

    if xml_text.startswith('<?xml '):
        if xml_declaration is False:
            lines = xml_text.splitlines()[1:]
        else:
            lines = xml_text.splitlines()
    elif xml_declaration and encoding.lower() != 'unicode':
        lines = ['<?xml version="1.0" encoding="{}"?>'.format(encoding)]
        lines.extend(xml_text.splitlines())
    else:
        lines = xml_text.splitlines()

    # Clear ending empty lines
    while lines and not lines[-1].strip():
        lines.pop(-1)

    if not lines or method == 'text' or (not indent and not max_lines):
        if encoding == 'unicode':
            return '\n'.join(lines)
        return '\n'.join(lines).encode(encoding)

    last_indent = ' ' * min(k for k in range(len(lines[-1])) if lines[-1][k] != ' ')
    if len(lines) > 2:
        try:
            child_indent = ' ' * min(
                k for line in lines[1:-1] for k in range(len(line)) if line[k] != ' '
            )
        except ValueError:
            child_indent = ''

        min_indent = min(child_indent, last_indent)
    else:
        min_indent = child_indent = last_indent

    start = len(min_indent) - len(indent)

    if max_lines is not None and len(lines) > max_lines + 2:
        lines = lines[:max_lines] + [child_indent + '...'] * 2 + lines[-1:]

    if encoding == 'unicode':
        return '\n'.join(reindent(line) for line in lines)
    return '\n'.join(reindent(line) for line in lines).encode(encoding)


__all__ = ['SafeExpatParser', 'defuse_xml', 'is_etree_element', 'is_lxml_etree_element',
           'is_etree_element_instance', 'is_etree_document', 'is_lxml_etree_document',
           'is_etree_document_instance', 'etree_iter_strings', 'etree_deep_equal',
           'etree_iter_paths', 'etree_tostring']



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from typing import Any, Optional, cast, TypeVar

from elementpath.aliases import XPathParserType
from elementpath.helpers import LazyPattern
from .any_types import AnyAtomicType
from .untyped import UntypedAtomic

__all__ = ['AbstractQName', 'QName', 'Notation']

T = TypeVar('T', bound='AbstractQName')


class AbstractQName(AnyAtomicType):
    """
    XPath compliant QName, bound with a prefix and a namespace.
    """
    pattern = LazyPattern(
        r'^(?:(?P<prefix>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*):)?'
        r'(?P<local>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*)$',
    )

    @classmethod
    def make(cls: type[T], value: Any,
             namespaces: dict[str, str] | None = None,
             parser: XPathParserType | None = None,
             **kwargs: Any) -> T:

        if isinstance(value, AbstractQName):
            return cast(T, value)
        elif isinstance(value, UntypedAtomic) and parser and parser.version >= '3.0':
            value = value.value
        elif not isinstance(value, str):
            raise cls._invalid_type(value)

        if namespaces is None:
            namespaces = parser.namespaces if parser is not None else {}

        if ':' not in value:
            return cls(namespaces.get(''), value)
        else:
            return cls(namespaces[value[0:value.index(':')]], value)

    __slots__ = ('uri', 'qname', 'prefix', 'local_name')

    def __new__(cls, *args: Any, **kwargs: Any) -> 'AbstractQName':
        if cls.__name__ == 'Notation':
            raise TypeError("can't instantiate xs:NOTATION objects")
        return super().__new__(cls)

    def __init__(self, uri: Optional[str], qname: str) -> None:
        """
        :param uri: the bound namespace URI, must be a not empty \
        URI if a prefixed name is provided for the 2nd argument.
        :param qname: a prefixed name or a local name.
        """
        if uri is None:
            self.uri = ''
        elif isinstance(uri, str):
            self.uri = uri
        else:
            raise TypeError('the 1st argument has an invalid type %r' % type(uri))

        if not isinstance(qname, str):
            raise TypeError('the 2nd argument has an invalid type %r' % type(qname))
        self.qname = qname.strip()

        match = self.pattern.match(self.qname)
        if match is None:
            raise ValueError('invalid value {!r} for an xs:QName'.format(self.qname))

        self.prefix = match.groupdict()['prefix']
        self.local_name = match.groupdict()['local']
        if not uri and self.prefix:
            msg = '{!r}: cannot associate a non-empty prefix with no namespace'
            raise ValueError(msg.format(self))

    @property
    def namespace(self) -> str:
        return self.uri

    @namespace.setter
    def namespace(self, uri: str) -> None:
        self.uri = uri

    @property
    def expanded_name(self) -> str:
        return '{%s}%s' % (self.uri, self.local_name) if self.uri else self.local_name

    @property
    def braced_uri_name(self) -> str:
        return 'Q{%s}%s' % (self.uri, self.local_name)

    def __repr__(self) -> str:
        return '%s(uri=%r, qname=%r)' % (self.__class__.__name__, self.uri, self.qname)

    def __str__(self) -> str:
        return self.qname

    def __hash__(self) -> int:
        return hash((self.uri, self.local_name))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, AbstractQName):
            return self.uri == other.uri and self.local_name == other.local_name
        elif isinstance(other, (str, UntypedAtomic)):
            return other == self.qname
        return NotImplemented


class QName(AbstractQName):
    name = 'QName'


class Notation(AbstractQName):
    name = 'NOTATION'



#
# Copyright (c), 2025-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""Type annotation aliases for elementpath."""
from collections.abc import MutableMapping, Iterator
from decimal import Decimal
from typing import Any, Optional, NoReturn, TYPE_CHECKING, TypeVar, Union

if TYPE_CHECKING:
    from .protocols import ElementProtocol, DocumentProtocol  # noqa: F401
    from .protocols import XsdElementProtocol, XsdAttributeProtocol  # noqa: F401
    from .protocols import DocumentType, ElementType, SchemaElemType  # noqa: F401
    from .datatypes import AnyAtomicType, AbstractDateTime, QName, AnyURI  # noqa: F401
    from .datatypes import Date, DateTime, Time, GregorianYear, GregorianMonth  # noqa: F401
    from .datatypes import GregorianDay, GregorianYearMonth, GregorianMonthDay  # noqa: F401
    from .datatypes import DateTimeStamp, Duration, UntypedAtomic  # noqa: F401
    from .datatypes import DayTimeDuration, YearMonthDuration  # noqa: F401
    from .schema_proxy import AbstractSchemaProxy  # noqa: F401
    from .xpath_nodes import XPathNode, ElementNode, AttributeNode  # noqa: F401
    from .xpath_nodes import TextNode, DocumentNode, NamespaceNode  # noqa: F401
    from .xpath_nodes import CommentNode, ProcessingInstructionNode  # noqa: F401
    from .xpath_tokens import XPathToken, XPathAxis, XPathFunction  # noqa: F401
    from .xpath_tokens import XPathConstructor, XPathMap, XPathArray  # noqa: F401
    from .xpath_context import XPathContext, XPathSchemaContext  # noqa: F401
    from .xpath1 import XPath1Parser  # noqa: F401
    from .xpath2 import XPath2Parser  # noqa: F401
    from .xpath30 import XPath30Parser  # noqa: F401
    from .xpath31 import XPath31Parser  # noqa: F401

###
# Namespace maps
NamespacesType = MutableMapping[str, str]
NsmapType = MutableMapping[Optional[str], str]  # compatible with the nsmap of lxml Element
AnyNsmapType = Union[NamespacesType, NsmapType, None]  # for composition and function arguments

###
# Datatypes
NumericType = Union[int, float, Decimal]
DateTimeType = Union[
    'DateTime', 'Date', 'Time', 'GregorianYear', 'GregorianMonth',
    'GregorianDay', 'GregorianYearMonth', 'GregorianMonthDay'
]
DurationType = Union['Duration', 'YearMonthDuration', 'DayTimeDuration']
ArithmeticType = Union[NumericType, DateTimeType, 'DurationType', 'UntypedAtomic']
AtomicType = Union['AnyAtomicType', str, bool, int, float, Decimal]

###
# XPath nodes
TypedNodeType = Union['AttributeNode', 'ElementNode']
TaggedNodeType = Union['ElementNode', 'CommentNode', 'ProcessingInstructionNode']
ElementMapType = dict[object, TaggedNodeType]
FindAttrType = Optional['XsdAttributeProtocol']
FindElemType = Optional['XsdElementProtocol']
RootNodeType = Union['DocumentNode', 'ElementNode']

ParentNodeType = Union['DocumentNode', 'ElementNode']
ChildNodeType = Union['TextNode', TaggedNodeType]

# All possible variants for node as item
XPathNodeType = Union['XPathNode', 'DocumentNode', 'NamespaceNode', 'AttributeNode',
                      'TextNode', 'ElementNode', 'CommentNode', 'ProcessingInstructionNode']

###
# Sequence and item/value types (specifics and generics)
ItemType = Union['XPathNode', AtomicType, 'XPathFunction']

_T = TypeVar('_T', bound=ItemType)

OneOrEmpty = Union[_T, list[_T], list[NoReturn]]
OneOrMore = Union[_T, list[_T]]
AnyOrEmpty = Union[_T, list[_T], list[NoReturn]]

SequenceArgType = Union[
    Iterator[_T], list[_T], list[NoReturn], tuple[_T, ...], tuple[_T], tuple[()]
]

OneAtomicOrEmpty = OneOrEmpty[AtomicType]
OneNumericOrEmpty = OneOrEmpty[NumericType]
OneArithmeticOrEmpty = OneOrEmpty[ArithmeticType]
OneItemOrEmpty = OneOrEmpty[ItemType]

OneOrMoreAtomic = OneOrMore[AtomicType]
OneOrMoreNumeric = OneOrMore[NumericType]
OneOrMoreArithmetic = OneOrMore[ArithmeticType]
OneOrMoreItems = OneOrMore[ItemType]

AnyAtomicOrEmpty = AnyOrEmpty[AtomicType]
AnyNumericOrEmpty = AnyOrEmpty[NumericType]
AnyArithmeticOrEmpty = AnyOrEmpty[ArithmeticType]
AnyItemsOrEmpty = AnyOrEmpty[ItemType]

ValueType = Union[ItemType, list[ItemType], list[NoReturn], list['XPathNode'],
                  list[Union['AttributeNode', 'ElementNode']], list[AtomicType]]
ListItemType = Union[ItemType, list[ItemType], list[NoReturn]]
ResultType = Union[
    AtomicType, 'ElementProtocol', 'XsdAttributeProtocol', tuple[Optional[str], str],
    'DocumentProtocol', 'DocumentNode', 'XPathFunction', object
]
MapDictType = dict[Optional[AtomicType], ValueType]
SequenceTypesType = Union[str, list[str], tuple[str, ...]]

# Parsers and tokens
XPathParserType = Union['XPath1Parser', 'XPath2Parser', 'XPath30Parser', 'XPath31Parser']
XPathTokenType = Union['XPathToken', 'XPathAxis', 'XPathFunction', 'XPathConstructor']
XPath2ParserType = Union['XPath2Parser', 'XPath30Parser', 'XPath31Parser']
ParserClassType = Union[
    type['XPath1Parser'], type['XPath2Parser'], type['XPath30Parser'], type['XPath31Parser']
]
NargsType = Optional[Union[int, tuple[int, Optional[int]]]]
ClassCheckType = Union[type[Any], tuple[type[Any], ...]]

# Context arguments
ContextType = Union['XPathContext', 'XPathSchemaContext', None]
RootArgType = Union['DocumentType', 'ElementType', 'SchemaElemType', RootNodeType]
ItemArgType = Union[ItemType, 'ElementProtocol', 'DocumentProtocol']
SchemaProxyType = Union['AbstractSchemaProxy']

_S = TypeVar('_S', bound=ItemArgType)

InputType = Union[None, _S, list[_S], list[NoReturn], tuple[_S, ...], tuple[_S], tuple[()]]
VariableValueType = InputType[ItemArgType]
FunctionArgType = Union[VariableValueType, ValueType]
NodeArgType = Union['XPathNode', 'ElementProtocol', 'DocumentProtocol']
CollectionArgType = InputType[NodeArgType]



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import math
from typing import Any, Optional, SupportsInt, Union

from elementpath.aliases import XPath2ParserType
from elementpath.helpers import FloatArgType, NUMERIC_INF_OR_NAN, INVALID_NUMERIC, \
    LazyPattern, collapse_white_spaces
from .any_types import AnyAtomicType

__all__ = ['Float', 'Float10', 'Integer', 'Int', 'Long',
           'NegativeInteger', 'PositiveInteger', 'NonNegativeInteger',
           'NonPositiveInteger', 'Short', 'Byte', 'UnsignedByte',
           'UnsignedInt', 'UnsignedLong', 'UnsignedShort']


class Float(float, AnyAtomicType):
    name = 'float'
    pattern = LazyPattern(
        r'^(?:[+-]?(?:[0-9]+(?:\.[0-9]*)?|\.[0-9]+)(?:[Ee][+-]?[0-9]+)? |[+-]?INF|NaN)$'
    )

    @classmethod
    def make(cls, value: Any,
             parser: XPath2ParserType | None = None,
             xsd_version: str | None = None,
             **kwargs: Any) -> 'Float':
        if cls.__name__.endswith('10'):
            return Float(value, '1.0')
        elif parser is not None:
            return Float(value, parser.xsd_version)
        else:
            return Float(value, xsd_version)

    def __new__(cls, value: FloatArgType | str, xsd_version: str | None = None) -> 'Float':
        if isinstance(value, str):
            value = collapse_white_spaces(value)
            if value in NUMERIC_INF_OR_NAN:
                if xsd_version == '1.0' and value == '+INF':
                    raise cls._invalid_value(value)
                elif value == 'NaN':
                    try:
                        return float_nan
                    except NameError:
                        pass
            elif value.lower() in INVALID_NUMERIC:
                raise cls._invalid_value(value)
        elif math.isnan(value):
            try:
                return float_nan
            except NameError:  # pragma: no cover
                pass

        _value = super().__new__(cls, value)
        if _value > 3.4028235E38:
            return super().__new__(cls, 'INF')
        elif _value < -3.4028235E38:
            return super().__new__(cls, '-INF')
        elif -1e-37 < _value < 1e-37:
            return super().__new__(cls, -0.0 if str(_value).startswith('-') else 0.0)
        return _value

    def __init__(self, value: FloatArgType, xsd_version: str | None = None) -> None:
        float.__init__(self)

    def __hash__(self) -> int:
        return super(Float, self).__hash__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            if super(Float, self).__eq__(other):
                return True
            return math.isclose(self, other, rel_tol=1e-7, abs_tol=0.0)
        return super(Float, self).__eq__(other)

    def __ne__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            if super(Float, self).__eq__(other):
                return False
            return not math.isclose(self, other, rel_tol=1e-7, abs_tol=0.0)
        return super(Float, self).__ne__(other)

    def __add__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__add__(other))
        elif isinstance(other, float):
            return super(Float, self).__add__(other)
        return NotImplemented

    def __radd__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__radd__(other))
        elif isinstance(other, float):
            return super(Float, self).__radd__(other)
        return NotImplemented

    def __sub__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__sub__(other))
        elif isinstance(other, float):
            return super(Float, self).__sub__(other)
        return NotImplemented

    def __rsub__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__rsub__(other))
        elif isinstance(other, float):
            return super(Float, self).__rsub__(other)
        return NotImplemented

    def __mul__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__mul__(other))
        elif isinstance(other, float):
            return super(Float, self).__mul__(other)
        return NotImplemented

    def __rmul__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__rmul__(other))
        elif isinstance(other, float):
            return super(Float, self).__rmul__(other)
        return NotImplemented

    def __truediv__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__truediv__(other))
        elif isinstance(other, float):
            return super(Float, self).__truediv__(other)
        return NotImplemented

    def __rtruediv__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__rtruediv__(other))
        elif isinstance(other, float):
            return super(Float, self).__rtruediv__(other)
        return NotImplemented

    def __mod__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__mod__(other))
        elif isinstance(other, float):
            return super(Float, self).__mod__(other)
        return NotImplemented

    def __rmod__(self, other: object) -> Union[float, 'Float']:
        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
            return self.__class__(super(Float, self).__rmod__(other))
        elif isinstance(other, float):
            return super(Float, self).__rmod__(other)
        return NotImplemented

    def __abs__(self) -> Union['Float']:
        return self.__class__(super(Float, self).__abs__())


class Float10(Float):
    """xs:float for XSD 1.0"""
    def __new__(cls, value: FloatArgType | str,  # type: ignore[misc]
                xsd_version: str | None = '1.0') -> 'Float':
        return Float(value, xsd_version=xsd_version)


# The instance used for xs:float NaN values in order to keep identity
float_nan = Float('NaN')


class Integer(int, AnyAtomicType):
    name = 'integer'
    pattern = LazyPattern(r'^[\-+]?[0-9]+$')

    _lower_bound: Optional[int] = None
    _higher_bound: Optional[int] = None

    def __init__(self, value: Union[str, SupportsInt]) -> None:
        """
        :param value: a string or an integer compatible value.
        """
        if self._lower_bound is not None and self < self._lower_bound:
            raise ValueError("value {} is too low for {!r}".format(value, self.__class__))
        elif self._higher_bound is not None and self >= self._higher_bound:
            raise ValueError("value {} is too high for {!r}".format(value, self.__class__))
        int.__init__(self)

    @classmethod
    def __subclasshook__(cls, subclass: type[Any]) -> bool:
        if cls is Integer:
            return issubclass(subclass, int) and not issubclass(subclass, bool)
        return NotImplemented  # type: ignore[no-any-return,unused-ignore]

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, cls):
            return
        elif isinstance(value, str):
            if cls.pattern.match(value) is None:
                raise cls._invalid_value(value)
        else:
            raise cls._invalid_type(value)


class NonPositiveInteger(Integer):
    name = 'nonPositiveInteger'
    _lower_bound, _higher_bound = None, 1


class NegativeInteger(NonPositiveInteger):
    name = 'negativeInteger'
    _lower_bound, _higher_bound = None, 0


class Long(Integer):
    name = 'long'
    _lower_bound, _higher_bound = -2 ** 63, 2 ** 63


class Int(Long):
    name = 'int'
    _lower_bound, _higher_bound = -2 ** 31, 2 ** 31


class Short(Int):
    name = 'short'
    _lower_bound, _higher_bound = -2 ** 15, 2 ** 15


class Byte(Short):
    name = 'byte'
    _lower_bound, _higher_bound = -2 ** 7, 2 ** 7


class NonNegativeInteger(Integer):
    name = 'nonNegativeInteger'
    _lower_bound = 0
    _higher_bound: Optional[int] = None


class PositiveInteger(NonNegativeInteger):
    name = 'positiveInteger'
    _lower_bound, _higher_bound = 1, None


class UnsignedLong(NonNegativeInteger):
    name = 'unsignedLong'
    _lower_bound, _higher_bound = 0, 2 ** 64


class UnsignedInt(UnsignedLong):
    name = 'unsignedInt'
    _lower_bound, _higher_bound = 0, 2 ** 32


class UnsignedShort(UnsignedInt):
    name = 'unsignedShort'
    _lower_bound, _higher_bound = 0, 2 ** 16


class UnsignedByte(UnsignedShort):
    name = 'unsignedByte'
    _lower_bound, _higher_bound = 0, 2 ** 8



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
Translation maps for XPath 3.0+ format functions. Add languages with pull-requests.
"""
from string import ascii_lowercase

ALPHABET_CHARACTERS = {
    None: ascii_lowercase,
    'en': ascii_lowercase,
    'it': 'abcdefghilmnopqrstuvz',
    'el': 'αβγδεζηθικλμνξοπρςστυφχψω',
}

OTHER_NUMBERS = (
    '\u2070\u00B9\u00B2\u00B3' +
    ''.join(chr(x) for x in range(0x2074, 0x207A)),  # superscript digits (0-9)
    ''.join(chr(x) for x in range(0x2080, 0x208A)),  # subscript digits (0-9)
    ''.join(chr(x) for x in range(0x2460, 0x2474)),  # circled numbers (1-20)
    ''.join(chr(x) for x in range(0x2474, 0x2488)),  # parenthesized numbers (1-20)
    ''.join(chr(x) for x in range(0x2488, 0x249C)),  # full stop numbers (1-20)
)

ROMAN_NUMERALS_MAP = {
    1000: 'M',
    900: 'CM',
    500: 'D',
    400: 'CD',
    100: 'C',
    90: 'XC',
    50: 'L',
    40: 'XL',
    10: 'X',
    9: 'IX',
    5: 'V',
    4: 'IV',
    1: 'I',
}

NUM_TO_MONTH_MAPS = {
    'en': {
        1: 'january',
        2: 'february',
        3: 'march',
        4: 'april',
        5: 'may',
        6: 'june',
        7: 'july',
        8: 'august',
        9: 'september',
        10: 'october',
        11: 'november',
        12: 'december',
    },
    'it': {
        1: 'gennaio',
        2: 'febbraio',
        3: 'marzo',
        4: 'aprile',
        5: 'maggio',
        6: 'giugno',
        7: 'luglio',
        8: 'agosto',
        9: 'settembre',
        10: 'ottobre',
        11: 'novembre',
        12: 'dicembre',
    },
}

NUM_TO_WEEKDAY_MAPS = {
    'en': {
        1: 'monday',
        2: 'tuesday',
        3: 'wednesday',
        4: 'thursday',
        5: 'friday',
        6: 'saturday',
        7: 'sunday',
    },
    'it': {
        1: 'lunedì',
        2: 'martedì',
        3: 'mercoledì',
        4: 'giovedì',
        5: 'venerdì',
        6: 'sabato',
        7: 'domenica',
    },
}


NUM_TO_WORD_MAPS = {
    'en': {
        10 ** 9: 'billion',
        10 ** 6: 'million',
        1000: 'thousand',
        100: 'hundred',
        90: 'ninety',
        80: 'eighty',
        70: 'seventy',
        60: 'sixty',
        50: 'fifty',
        40: 'forty',
        30: 'thirty',
        20: 'twenty',
        19: 'nineteen',
        18: 'eighteen',
        17: 'seventeen',
        16: 'sixteen',
        15: 'fifteen',
        14: 'fourteen',
        13: 'thirteen',
        12: 'twelve',
        11: 'eleven',
        10: 'ten',
        9: 'nine',
        8: 'eight',
        7: 'seven',
        6: 'six',
        5: 'five',
        4: 'four',
        3: 'three',
        2: 'two',
        1: 'one',
        0: 'zero',
    },
    'it': {
        10 ** 9: 'miliardo',
        10 ** 6: 'milione',
        1000: 'mille',
        100: 'cento',
        90: 'novanta',
        80: 'ottanta',
        70: 'settanta',
        60: 'sessanta',
        50: 'cinquanta',
        40: 'quaranta',
        30: 'trenta',
        20: 'venti',
        19: 'diciannove',
        18: 'diciotto',
        17: 'diciassette',
        16: 'sedici',
        15: 'quindici',
        14: 'quattordici',
        13: 'tredici',
        12: 'dodici',
        11: 'undici',
        10: 'dieci',
        9: 'nove',
        8: 'otto',
        7: 'sette',
        6: 'sei',
        5: 'cinque',
        4: 'quattro',
        3: 'tre',
        2: 'due',
        1: 'uno',
        0: 'zero',
    }
}

MILITARY_TIME_ZONES = {
    '+01': 'A',
    '+02': 'B',
    '+03': 'C',
    '+04': 'D',
    '+05': 'E',
    '+06': 'F',
    '+07': 'G',
    '+08': 'H',
    '+09': 'I',
    None: 'J',
    '+10': 'K',
    '+11': 'L',
    '+12': 'M',
    '-01': 'N',
    '-02': 'O',
    '-03': 'P',
    '-04': 'Q',
    '-05': 'R',
    '-06': 'S',
    '-07': 'T',
    '-08': 'U',
    '-09': 'V',
    '-10': 'W',
    '-11': 'X',
    '-12': 'Y',
    '+00': 'Z',
}



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 1.0 implementation - part 1 (parser class and symbols)
"""
import re
from abc import ABCMeta
from collections.abc import Callable, MutableMapping, Sequence
from typing import cast, Any, ClassVar, Optional, Union

import elementpath.aliases as ta

from elementpath.exceptions import xpath_error, UnsupportedFeatureError, \
    ElementPathValueError, ElementPathNameError, ElementPathKeyError, MissingContextError
from elementpath.namespaces import XML_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE
from elementpath.helpers import upper_camel_case
from elementpath.collations import UNICODE_CODEPOINT_COLLATION
from elementpath.datatypes import QName
from elementpath.tdop import Parser
from elementpath.sequence_types import get_function_signatures, match_sequence_type
from elementpath.schema_proxy import AbstractSchemaProxy
from elementpath.xpath_tokens import XPathToken, XPathAxis, XPathFunction, ProxyToken, \
    NameToken, PrefixedNameToken, BracedNameToken


class XPath1Parser(Parser[ta.XPathTokenType]):
    """
    XPath 1.0 expression parser class. Provide a *namespaces* dictionary argument for
    mapping namespace prefixes to URI inside expressions. If *strict* is set to `False`
    the parser enables also the parsing of QNames, like the ElementPath library.

    :param namespaces: a dictionary with mapping from namespace prefixes into URIs.
    :param strict: a strict mode is `False` the parser enables parsing of QNames \
    in extended format, like the Python's ElementPath library. Default is `True`.
    """
    version = '1.0'
    """The XPath version string."""

    token_base_class = XPathToken  # type: ignore[assignment, unused-ignore]
    literals_pattern = re.compile(
        r"""'(?:[^']|'')*'|"(?:[^"]|"")*"|(?:\d+|\.\d+)(?:\.\d*)?(?:[Ee][+-]?\d+)?"""
    )
    name_pattern = re.compile(r'[^\d\W][\w.\-\xb7\u0300-\u036F\u203F\u2040]*')

    RESERVED_FUNCTION_NAMES = {
        'comment', 'element', 'node', 'processing-instruction', 'text'
    }

    DEFAULT_NAMESPACES: ClassVar[dict[str, str]] = {'xml': XML_NAMESPACE}
    """Namespaces known statically by default."""

    # Labels and symbols admitted after a path step
    PATH_STEP_LABELS: ClassVar[tuple[str, ...]] = ('axis', 'kind test')
    PATH_STEP_SYMBOLS: ClassVar[set[str]] = {
        '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '{'
    }

    # Class attributes for compatibility with XPath 2.0+
    schema: Optional[AbstractSchemaProxy] = None
    variable_types: Optional[dict[str, str]] = None
    document_types: Optional[dict[str, str]] = None
    collection_types: Optional[ta.NamespacesType] = None
    default_collection_type: str = 'node()*'
    base_uri: Optional[str] = None
    function_namespace = XPATH_FUNCTIONS_NAMESPACE
    function_signatures: dict[tuple[QName, int], str] = {}
    decimal_formats: dict[Optional[str], Any] = {}
    parse_arguments: bool = True
    defuse_xml: bool = True

    compatibility_mode: bool = True
    """XPath 1.0 compatibility mode."""

    default_namespace: Optional[str] = None
    """
    The default namespace. For XPath 1.0 this value is always `None` because the default
    namespace is ignored (see https://www.w3.org/TR/1999/REC-xpath-19991116/#node-tests).
    """

    default_collation = UNICODE_CODEPOINT_COLLATION

    @staticmethod
    def tracer(trace_data: str) -> None:
        """Trace data collector"""

    def __init__(self, namespaces: Optional[ta.NamespacesType] = None,
                 strict: bool = True) -> None:
        super(XPath1Parser, self).__init__()
        self.namespaces: dict[str, str] = self.DEFAULT_NAMESPACES.copy()
        if namespaces is not None:
            self.namespaces.update(namespaces)
        self.strict: bool = strict

    def __str__(self) -> str:
        args = []
        if self.namespaces != self.DEFAULT_NAMESPACES:
            args.append(str(self.other_namespaces))
        if not self.strict:
            args.append('strict=False')
        return f"{self.__class__.__name__}({', '.join(args)})"

    @property
    def other_namespaces(self) -> dict[str, str]:
        """The subset of namespaces not known by default."""
        return {k: v for k, v in self.namespaces.items()
                if k not in self.DEFAULT_NAMESPACES or self.DEFAULT_NAMESPACES[k] != v}

    @property
    def xsd_version(self) -> str:
        return '1.0'  # Use XSD 1.0 datatypes for default

    def is_schema_bound(self) -> bool:
        return False

    def xsd_qname(self, local_name: str) -> str:
        """Returns a prefixed QName string for XSD namespace."""
        if self.namespaces.get('xs') == XSD_NAMESPACE:
            return 'xs:%s' % local_name

        for pfx, uri in self.namespaces.items():
            if uri == XSD_NAMESPACE:
                return '%s:%s' % (pfx, local_name) if pfx else local_name

        raise xpath_error('XPST0081', 'Missing XSD namespace registration')

    @classmethod
    def create_restricted_parser(cls, name: str, symbols: Sequence[str]) \
            -> type['XPath1Parser']:
        """Get a parser subclass with a restricted set of symbols.s"""
        symbol_table = {
            k: v for k, v in cls.symbol_table.items() if k in symbols
        }
        return cast(type['XPath1Parser'], ABCMeta(
            f"{name}{cls.__name__}", (cls,), {'symbol_table': symbol_table}
        ))

    @staticmethod
    def unescape(string_literal: str) -> str:
        if string_literal.startswith("'"):
            return string_literal[1:-1].replace("''", "'")
        else:
            return string_literal[1:-1].replace('""', '"')

    @classmethod
    def proxy(cls, symbol: str, label: str = 'proxy', bp: int = 90) -> type[ProxyToken]:
        """Register a proxy token class for a symbol."""
        if symbol in cls.symbol_table and not issubclass(cls.symbol_table[symbol], ProxyToken):
            # Move the token class before register the proxy token
            token_cls = cls.symbol_table.pop(symbol)
            cls.symbol_table[f'{{{token_cls.namespace}}}{symbol}'] = token_cls

        token_class_name = "_%s%sProxy" % (
            upper_camel_case(symbol), str(label).title().replace(' ', '')
        )
        token_class = cls.register(
            symbol,
            label='function',
            class_name=token_class_name,
            bases=(ProxyToken,),
            lbp=bp,
            rbp=bp
        )
        assert issubclass(token_class, ProxyToken)
        return token_class

    @classmethod
    def axis(cls, symbol: str, reverse_axis: bool = False, bp: int = 80) -> type[XPathAxis]:
        """Register a token class for a symbol that represents an XPath *axis*."""
        token_class = cls.register(symbol, bases=(XPathAxis,),
                                   reverse_axis=reverse_axis, lbp=bp, rbp=bp)
        assert issubclass(token_class, XPathAxis)
        return token_class

    @classmethod
    def function(cls, symbol: str,
                 prefix: Optional[str] = None,
                 label: str = 'function',
                 nargs: ta.NargsType = None,
                 sequence_types: tuple[str, ...] = (),
                 bp: int = 90) -> type[XPathFunction]:
        """
        Registers a token class for a symbol that represents an XPath function.
        """
        kwargs: dict[str, Any] = {
            'bases': (XPathFunction,),
            'label': label,
            'nargs': nargs,
            'lbp': bp,
            'rbp': bp,
        }
        if 'function' not in label:
            # kind test or sequence type
            return cast(type[XPathFunction], cls.register(symbol, **kwargs))
        elif symbol in cls.RESERVED_FUNCTION_NAMES:
            raise ElementPathValueError(f'{symbol!r} is a reserved function name')

        if prefix:
            namespace = cls.DEFAULT_NAMESPACES[prefix]
            qname = QName(namespace, '%s:%s' % (prefix, symbol))
            kwargs['lookup_name'] = qname.expanded_name
            kwargs['class_name'] = '_%s%s%s' % (
                prefix.capitalize(),
                symbol.capitalize(),
                str(label).title().replace(' ', '')
            )
            kwargs['namespace'] = namespace
            cls.proxy(symbol, label='function', bp=bp)
        else:
            qname = QName(XPATH_FUNCTIONS_NAMESPACE, 'fn:%s' % symbol)
            kwargs['namespace'] = XPATH_FUNCTIONS_NAMESPACE

        if sequence_types:
            # Register function signature(s)
            cls.function_signatures.update(
                get_function_signatures(qname, nargs, sequence_types)
            )
            kwargs['sequence_types'] = sequence_types

        return cast(type[XPathFunction], cls.register(symbol, **kwargs))

    def parse(self, source: str) -> XPathToken:
        if self.tokenizer is None:
            self.tokenizer = self.create_tokenizer(self.symbol_table)

        root_token = super().parse(source)
        if root_token.label in ('sequence type', 'function test'):
            raise root_token.error('XPST0003', "not allowed in XPath expression")

        try:
            root_token.evaluate()  # Static context evaluation
        except MissingContextError:
            pass

        if self.schema is not None:
            # Static evaluation using a schema context
            context = self.schema.get_context()
            for _ in root_token.select(context):
                pass

        return root_token

    def expected_next(self, *symbols: str, message: Optional[str] = None) -> None:
        """
        Checks the next token with a list of symbols. Replaces the next token with
        a '(name)' token if the check fails and the next token can be a name,
        otherwise raises a syntax error.

        :param symbols: a sequence of symbols.
        :param message: optional error message.
        """
        if self.next_token.symbol in symbols:
            return
        elif '(name)' in symbols and \
                not isinstance(self.next_token, (XPathFunction, XPathAxis)) and \
                self.name_pattern.match(self.next_token.symbol) is not None:
            # Disambiguation replacing the next token with a '(name)' token
            cls = self.symbol_table['(name)']
            assert not issubclass(cls, XPathFunction)
            value = self.next_token.symbol
            self.next_token = cls(self, value)
        else:
            raise self.next_token.wrong_syntax(message)

    def parse_occurrence(self, token: XPathToken) -> None:
        """Parse the occurrence for the current token."""
        if self.next_token.symbol in ('*', '+', '?'):
            assert self.token is token
            token.occurrence = self.next_token.symbol
            self.advance()
            self.next_token.unexpected('*', '+', '?')

    def parse_sequence_type(self) -> XPathToken:
        if self.next_token.label in ('kind test', 'sequence type', 'function test'):
            token = self.expression(rbp=85)
        else:
            if self.next_token.symbol == 'Q{':
                token = self.advance().nud()
            elif self.next_token.symbol != '(name)':
                raise self.next_token.wrong_syntax()
            else:
                self.advance()
                if self.next_token.symbol == ':':
                    left = self.token
                    self.advance()
                    token = self.token.led(left)
                else:
                    token = self.token

                if self.next_token.symbol in ('::', '('):
                    raise self.next_token.wrong_syntax()

        next_symbol = self.next_token.symbol
        if token.symbol != 'empty-sequence' and next_symbol in ('?', '*', '+'):
            token.occurrence = next_symbol
            self.advance()
        return token

    def check_variables(self, values: MutableMapping[str, Any]) -> None:
        """Checks the sequence types of the XPath dynamic context's variables."""
        for varname, value in values.items():
            if not match_sequence_type(
                value, 'item()*' if isinstance(value, list) else 'item()', self
            ):
                message = "Unmatched sequence type for variable {!r}".format(varname)
                raise xpath_error('XPDY0050', message)

    def get_function(self, name: str, arity: Optional[int],
                     context: ta.ContextType = None) -> XPathFunction:
        """
        Returns an XPathFunction object suitable for stand-alone usage.

        :param name: the name of the function.
        :param arity: the arity of the function object, must be compatible \
        with the signature of the XPath function.
        :param context: an optional context to bound to the function.
        """
        if ':' not in name:
            qname = QName(XPATH_FUNCTIONS_NAMESPACE, f'fn:{name}')
        elif name.startswith('fn:'):
            qname = QName(XPATH_FUNCTIONS_NAMESPACE, name)
            name = name[3:]
        else:
            prefix, name = name.split(':')
            try:
                namespace = self.namespaces[prefix]
            except KeyError:
                raise ElementPathKeyError(f"Unknown namespace {prefix!r}") from None
            else:
                qname = QName(namespace, f'{prefix}:{name}')

        if qname.expanded_name in self.symbol_table:
            token_class = self.symbol_table[qname.expanded_name]
        elif name in self.symbol_table:
            token_class = self.symbol_table[name]
        else:
            raise ElementPathNameError(f'unknown function {name!r}')

        if not issubclass(token_class, XPathFunction):
            raise ElementPathNameError(f'{name!r} is not an XPath function')

        if token_class.namespace != qname.namespace:
            raise ElementPathNameError(f'namespace mismatch: {token_class.namespace}')

        try:
            func = token_class(self, nargs=arity)
        except TypeError:
            msg = f"unknown function {qname.qname}#{arity}"
            raise xpath_error('XPST0017', msg) from None
        else:
            if context is not None:
                func.context = context
            return func

    ###
    # Unsupported methods in XPath 1.0
    @classmethod
    def constructor(cls, symbol: str, bp: int = 90, nargs: ta.NargsType = 1,
                    sequence_types: Union[tuple[()], tuple[str, ...], list[str]] = (),
                    label: Union[str, tuple[str, ...]] = 'constructor function') \
            -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """
        Statically creates a constructor token class, that is registered in the globals
        of the module where the method is called.
        """
        raise UnsupportedFeatureError("Static definition of schema constructors token "
                                      "classes requires an XPath 2.0+ parser")

    def schema_constructor(self, atomic_type_name: str, bp: int = 90) \
            -> type[XPathFunction]:
        """Dynamically registers a token class for a schema atomic type constructor function."""
        raise UnsupportedFeatureError("Dynamic definition of schema constructors token "
                                      "classes requires an XPath 2.0+ parser")

    def external_function(self,
                          callback: Callable[..., Any],
                          name: Optional[str] = None,
                          prefix: Optional[str] = None,
                          sequence_types: tuple[str, ...] = (),
                          bp: int = 90) -> type[XPathFunction]:
        """Registers a token class for an external function."""
        raise UnsupportedFeatureError(
            "Registration of external functions requires an XPath 2.0+ parser"
        )


###
# Special symbols
XPath1Parser.register('(start)')
XPath1Parser.register('(end)')
XPath1Parser.register('(invalid)')
XPath1Parser.register('(unknown)')
XPath1Parser.literal('(string)')
XPath1Parser.literal('(float)')
XPath1Parser.literal('(decimal)')
XPath1Parser.literal('(integer)')

###
# Simple symbols
XPath1Parser.register(',')
XPath1Parser.register(')', bp=100)
XPath1Parser.register(']')
XPath1Parser.register('::')
XPath1Parser.register('}')

###
# Name related tokens
XPath1Parser.symbol_table['(name)'] = NameToken
XPath1Parser.symbol_table[':'] = PrefixedNameToken
XPath1Parser.symbol_table['{'] = BracedNameToken

# XPath 1.0 definitions continue into module _xpath1_operators



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import operator
from collections.abc import Callable
from decimal import Decimal
from typing import Any, cast, Union

from elementpath.aliases import XPath2ParserType
from elementpath.helpers import reversed_sub, reversed_truediv, BOOLEAN_VALUES, get_double
from .any_types import AnyAtomicType

UntypedArgType = Union[str, bytes, bool, float, Decimal, 'UntypedAtomic', AnyAtomicType]

__all__ = ['UntypedAtomic']


class UntypedAtomic(AnyAtomicType):
    """
    Class for xs:untypedAtomic data. Provides special methods for comparing
    and converting to basic data types.

    :param value: the untyped value, usually a string.
    :param parser: the XPath parser that creates the instance, if any.
    """
    name = 'untypedAtomic'
    value: str

    __slots__ = ('value', '_xsd_version', 'parser')

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'UntypedAtomic':
        return cls(value, kwargs.get('parser'))

    @classmethod
    def validate(cls, value: object) -> None:
        if not isinstance(value, cls):
            raise cls._invalid_type(value)

    def __init__(self, value: UntypedArgType,
                 parser: XPath2ParserType | None = None) -> None:
        match value:
            case str():
                self.value = value
            case bytes():
                self.value = value.decode('utf-8')
            case bool():
                self.value = 'true' if value else 'false'
            case float():
                self.value = str(value).rstrip('0').rstrip('.')
            case Decimal():
                self.value = str(value.normalize())
            case UntypedAtomic():
                self.value = value.value
            case None:
                raise TypeError("{!r} is not an atomic value".format(value))
            case AnyAtomicType():
                self.value = str(value)
            case _:
                raise TypeError("{!r} is not an atomic value".format(value))

        self.parser = parser
        self._xsd_version = parser.xsd_version if parser is not None else None

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, self.value)

    def _operator(self, op: Callable[[Any, Any], bool],
                  other: object, force_float: bool = True) -> bool:
        """
        Returns a couple of operands, applying a cast to the instance value based on
        the type of the *other* argument.

        :param other: The other operand, that determines the cast for the untyped instance.
        :param force_float: Force a conversion to float if *other* is an UntypedAtomic instance.
        :return: A couple of values.
        """
        match other:
            case UntypedAtomic():
                if force_float:
                    return op(get_double(self.value, self._xsd_version),
                              get_double(other.value, self._xsd_version))
                return op(self.value, other.value)
            case bool():
                # Cast to xs:boolean
                value = self.value.strip()
                if value not in BOOLEAN_VALUES:
                    raise ValueError("{!r} cannot be cast to xs:boolean".format(self.value))
                return op(value in ('1', 'true'), other)
            case int():
                return op(get_double(self.value, self._xsd_version), other)
            case None | str() | list():
                return op(self.value, other)
            case AnyAtomicType():
                if hasattr(other, 'make'):
                    return op(type(other).make(self.value, parser=self.parser), other)
                else:
                    return op(type(other)(self.value), other)
            case _:
                return cast(bool, NotImplemented)

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other: Any) -> bool:
        return self._operator(operator.eq, other, force_float=False)

    def __ne__(self, other: Any) -> bool:
        return not self._operator(operator.eq, other, force_float=False)

    def __lt__(self, other: Any) -> bool:
        return self._operator(operator.lt, other)

    def __le__(self, other: Any) -> bool:
        return self._operator(operator.le, other)

    def __gt__(self, other: Any) -> bool:
        return self._operator(operator.gt, other)

    def __ge__(self, other: Any) -> bool:
        return self._operator(operator.ge, other)

    def __add__(self, other: Any) -> bool:
        return self._operator(operator.add, other)
    __radd__ = __add__

    def __sub__(self, other: Any) -> Any:
        return self._operator(operator.sub, other)

    def __rsub__(self, other: Any) -> Any:
        return self._operator(reversed_sub, other)

    def __mul__(self, other: Any) -> Any:
        return self._operator(operator.mul, other)
    __rmul__ = __mul__

    def __truediv__(self, other: Any) -> Any:
        return self._operator(operator.truediv, other)

    def __rtruediv__(self, other: Any) -> Any:
        return self._operator(reversed_truediv, other)

    def __int__(self) -> int:
        return int(self.value)

    def __float__(self) -> float:
        return get_double(self.value, self._xsd_version)

    def __bool__(self) -> bool:
        return bool(self.value)  # For effective boolean value, not for cast to xs:boolean.

    def __abs__(self) -> Decimal:
        return abs(Decimal(self.value))

    def __mod__(self, other: Any) -> Any:
        return self._operator(operator.mod, other)

    def __round__(self, n: int | None = None) -> float:
        return round(float(self.value), ndigits=n)

    def __str__(self) -> str:
        return self.value

    def __bytes__(self) -> bytes:
        return bytes(self.value, encoding='utf-8')



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from abc import ABCMeta, abstractmethod
from collections.abc import Iterator
from functools import lru_cache
from typing import Any, Optional, Union

from elementpath.exceptions import ElementPathTypeError
from elementpath.protocols import XsdTypeProtocol, XsdAttributeProtocol, \
    XsdElementProtocol, XsdSchemaProtocol
from elementpath.aliases import AtomicType, XPath2ParserType
from elementpath.etree import is_etree_element
from elementpath.xpath_context import XPathSchemaContext

PathResult = Union[XsdSchemaProtocol, XsdElementProtocol, XsdAttributeProtocol]
FindAttrType = Optional[XsdAttributeProtocol]
FindElemType = Optional[XsdElementProtocol]


class AbstractSchemaProxy(metaclass=ABCMeta):
    """
    Abstract base class for defining schema proxies. An implementation can override
    initialization type annotations

    :param schema: a schema instance compatible with the XsdSchemaProtocol.
    :param base_element: the schema element used as base item for static analysis.
    """
    __slots__ = ('_schema', '_base_element', '_is_fully_valid')

    def __init__(self, schema: XsdSchemaProtocol,
                 base_element: Optional[XsdElementProtocol] = None) -> None:
        if not is_etree_element(schema):
            raise ElementPathTypeError(
                "argument {!r} is not a compatible schema instance".format(schema)
            )
        if base_element is not None and not is_etree_element(base_element):
            raise ElementPathTypeError(
                "argument 'base_element' is not a compatible element instance"
            )

        self._schema = schema
        self._base_element: Optional[XsdElementProtocol] = base_element
        self._is_fully_valid = False

    @property
    def schema(self) -> XsdSchemaProtocol:
        return self._schema

    @property
    def base_element(self) -> Optional[XsdElementProtocol]:
        return self._base_element

    @property
    def validity(self) -> str:
        validity = self._schema.validity
        if validity != 'valid':
            self._is_fully_valid = False
        return validity

    @property
    def validation_attempted(self) -> str:
        validation_attempted = self._schema.validation_attempted
        if validation_attempted != 'full':
            self._is_fully_valid = False
        return validation_attempted

    def is_fully_valid(self) -> bool:
        if self._is_fully_valid:
            return True
        self._is_fully_valid = self.validity == 'valid' and self.validation_attempted == 'full'
        return self._is_fully_valid

    def bind_parser(self, parser: XPath2ParserType) -> None:
        """
        Binds a parser instance with schema proxy adding the schema's atomic types constructors.
        This method can be redefined in a concrete proxy to optimize schema bindings.

        :param parser: a parser instance.
        """
        if parser.schema is not self:
            parser.schema = self

        for xsd_type in self.iter_atomic_types():
            if xsd_type.name is not None:  # pragma: no cover
                parser.schema_constructor(xsd_type.name)

    def get_context(self) -> XPathSchemaContext:
        """
        Get a context instance for static analysis phase.

        :returns: an `XPathSchemaContext` instance.
        """
        return XPathSchemaContext(root=self._schema, item=self._base_element, schema=self)

    def find(self, path: str, namespaces: Optional[dict[str, str]] = None) \
            -> Optional[PathResult]:
        """
        Find a schema element or attribute using an XPath expression. Currently unused
        in the code, it may be deprecated in the future.

        :param path: an XPath expression that selects an element or an attribute node.
        :param namespaces: an optional mapping from namespace prefix to namespace URI.
        :return: The first matching schema component, or ``None`` if there is no match.
        """
        if self._base_element is not None:
            return self._base_element.find(path, namespaces=namespaces)
        return self._schema.find(path, namespaces)

    @lru_cache(maxsize=None)
    def cached_find(self, expanded_path: str) -> Optional[PathResult]:
        """
        Find a schema element or attribute using an expanded XPath expression.
        Currently unused in the code, it may be deprecated in the future.
        """
        if self._base_element is not None:
            return self._base_element.find(expanded_path)
        return self._schema.find(expanded_path)

    @property
    def xsd_version(self) -> str:
        """The XSD version, returns '1.0' or '1.1'."""
        return self._schema.xsd_version

    def is_assertion_based(self) -> bool:
        return self._base_element is not None and \
            self._base_element.parent is self._base_element.type

    def get_type(self, qname: str) -> Optional[XsdTypeProtocol]:
        """
        Get the XSD global type from the schema's scope.

        :param qname: the fully qualified name of the type to retrieve.
        :returns: an object that represents an XSD type or `None`.
        """
        return self._schema.maps.types.get(qname)

    def get_attribute(self, qname: str) -> Optional[XsdAttributeProtocol]:
        """
        Get the XSD global attribute from the schema's scope.

        :param qname: the fully qualified name of the attribute to retrieve.
        """
        return self._schema.maps.attributes.get(qname)

    def get_element(self, qname: str,) -> Optional[XsdElementProtocol]:
        """
        Get the XSD global element from the schema's scope.

        :param qname: the fully qualified name of the attribute to retrieve.
        """
        return self._schema.maps.elements.get(qname)

    def get_substitution_group(self, qname: str) -> Optional[set[XsdElementProtocol]]:
        """
        Get a substitution group. Currently unused in the code, it may be deprecated
        in the future.

        :param qname: the fully qualified name of the substitution group to retrieve.
        :returns: a list containing substitution elements or `None`.
        """
        return self._schema.maps.substitution_groups.get(qname)

    @abstractmethod
    def is_instance(self, obj: Any, type_qname: str) -> bool:
        """
        Returns `True` if *obj* is an instance of the XSD global type, `False` if not.

        :param obj: the instance to be tested.
        :param type_qname: the fully qualified name of the type used to test the instance.
        """

    @abstractmethod
    def cast_as(self, obj: Any, type_qname: str) -> AtomicType:
        """
        Converts *obj* to the Python type associated with an XSD global type. A concrete
        implementation must raises a `ValueError` or `TypeError` in case of a decoding
        error or a `KeyError` if the type is not bound to the schema's scope.

        :param obj: the instance to be cast.
        :param type_qname: the fully qualified name of the type used to convert the instance.
        """

    @abstractmethod
    def iter_atomic_types(self) -> Iterator[XsdTypeProtocol]:
        """
        Returns an iterator for not builtin atomic types defined in the schema's scope. A concrete
        implementation must yield objects that implement the protocol `XsdTypeProtocol`.
        """


__all__ = ['PathResult', 'AbstractSchemaProxy']



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 1.0 implementation - part 2 (operators and expressions)
"""
import math
import decimal
import operator
from collections.abc import Iterator
from copy import copy
from typing import cast

import elementpath.aliases as ta

from elementpath.datatypes import AbstractDateTime, ArithmeticProxy, Duration, NumericProxy
from elementpath.xpath_nodes import XPathNode, ElementNode, DocumentNode

from elementpath.exceptions import ElementPathTypeError
from elementpath.helpers import node_position
from elementpath.xpath_context import XPathSchemaContext
from elementpath.xpath_tokens import XPathToken, NameToken, VariableToken, \
    ContextItemToken, AsteriskToken, ParentShortcutToken

from .xpath1_parser import XPath1Parser

__all__ = ['XPath1Parser']

OPERATORS_MAP = {
    '=': operator.eq,
    '!=': operator.ne,
    '>': operator.gt,
    '>=': operator.ge,
    '<': operator.lt,
    '<=': operator.le,
}

register = XPath1Parser.register
nullary = XPath1Parser.nullary
infix = XPath1Parser.infix
method = XPath1Parser.method

XPath1Parser.symbol_table['$'] = VariableToken
XPath1Parser.symbol_table['*'] = AsteriskToken
XPath1Parser.symbol_table['.'] = ContextItemToken
XPath1Parser.symbol_table['..'] = ParentShortcutToken


###
# Logical Operators
@method(infix('or', bp=20))
def evaluate__or_operator(self: XPathToken, context: ta.ContextType = None) -> bool:
    return self.boolean_value(self[0].select(copy(context))) or \
        self.boolean_value(self[1].select(copy(context)))


@method(infix('and', bp=25))
def evaluate__and_operator(self: XPathToken, context: ta.ContextType = None) -> bool:
    return self.boolean_value(self[0].select(copy(context))) and \
        self.boolean_value(self[1].select(copy(context)))


###
# Comparison operators
@method('=', bp=30)
@method('!=', bp=30)
@method('<', bp=30)
@method('>', bp=30)
@method('<=', bp=30)
@method('>=', bp=30)
def led__comparison_operators(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol in OPERATORS_MAP:
        raise self.wrong_syntax()
    self[:] = left, self.parser.expression(rbp=30)
    return self


@method('=')
@method('!=')
@method('<')
@method('>')
@method('<=')
@method('>=')
def evaluate__comparison_operators(self: XPathToken, context: ta.ContextType = None) -> bool:
    op = OPERATORS_MAP[self.symbol]
    try:
        return any(op(x1, x2) for x1, x2 in self.iter_comparison_data(context))
    except (TypeError, ValueError) as err:
        if isinstance(context, XPathSchemaContext):
            return False
        elif isinstance(err, ElementPathTypeError):
            raise
        elif isinstance(err, TypeError):
            raise self.error('XPTY0004', err) from None
        else:
            raise self.error('FORG0001', err) from None


###
# Numerical operators
@method(infix('+', bp=40))
def evaluate__plus_operator(self: XPathToken, context: ta.ContextType = None) \
        -> ta.OneArithmeticOrEmpty:
    if len(self) == 1:
        arg: ta.NumericType = self.get_argument(context, cls=NumericProxy)
        return [] if arg is None else +arg
    else:
        op1: ta.ArithmeticType | None
        op2: ta.ArithmeticType
        op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
        if op1 is None:
            return []

        try:
            return op1 + op2  # type:ignore[operator, return-value]
        except (TypeError, OverflowError) as err:
            if isinstance(context, XPathSchemaContext):
                return []
            elif isinstance(err, TypeError):
                raise self.error('XPTY0004', err) from None
            elif isinstance(op1, AbstractDateTime):
                raise self.error('FODT0001', err) from None
            elif isinstance(op1, Duration):
                raise self.error('FODT0002', err) from None
            else:
                raise self.error('FOAR0002', err) from None


@method(infix('-', bp=40))
def evaluate__minus_operator(self: XPathToken, context: ta.ContextType = None) \
        -> ta.OneArithmeticOrEmpty:
    if len(self) == 1:
        arg: ta.NumericType = self.get_argument(context, cls=NumericProxy)
        return [] if arg is None else -arg
    else:
        op1: ta.ArithmeticType | None
        op2: ta.ArithmeticType
        op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
        if op1 is None:
            return []

        try:
            return op1 - op2  # type:ignore[operator, return-value]
        except (TypeError, OverflowError) as err:
            if isinstance(context, XPathSchemaContext):
                return []
            elif isinstance(err, TypeError):
                raise self.error('XPTY0004', err) from None
            elif isinstance(op1, AbstractDateTime):
                raise self.error('FODT0001', err) from None
            elif isinstance(op1, Duration):
                raise self.error('FODT0002', err) from None
            else:
                raise self.error('FOAR0002', err) from None


@method('+')
@method('-')
def nud__plus_minus_operators(self: XPathToken) -> XPathToken:
    self[:] = self.parser.expression(rbp=70),
    return self


@method(infix('div', bp=45))
def evaluate__div_operator(self: XPathToken, context: ta.ContextType = None) \
        -> int | float | decimal.Decimal | ta.AnyItemsOrEmpty:
    dividend: ta.ArithmeticType | None
    divisor: ta.ArithmeticType
    dividend, divisor = self.get_operands(context, cls=ArithmeticProxy)
    if dividend is None:
        return []
    elif divisor != 0:
        try:
            if isinstance(dividend, int) and isinstance(divisor, int):
                return decimal.Decimal(dividend) / decimal.Decimal(divisor)
            return dividend / divisor  # type:ignore[operator]
        except TypeError as err:
            raise self.error('XPTY0004', err) from None
        except ValueError as err:
            raise self.error('FOCA0005', err) from None
        except OverflowError as err:
            raise self.error('FOAR0002', err) from None
        except (ZeroDivisionError, decimal.DivisionByZero):
            raise self.error('FOAR0001') from None

    elif isinstance(dividend, AbstractDateTime):
        raise self.error('FODT0001')
    elif isinstance(dividend, Duration):
        raise self.error('FODT0002')
    elif not self.parser.compatibility_mode and \
            isinstance(dividend, (int, decimal.Decimal)) and \
            isinstance(divisor, (int, decimal.Decimal)):
        raise self.error('FOAR0001')
    elif dividend == 0:
        return math.nan
    elif dividend > 0:
        return float('-inf') if str(divisor).startswith('-') else float('inf')
    else:
        return float('inf') if str(divisor).startswith('-') else float('-inf')


@method(infix('mod', bp=45))
def evaluate__mod_operator(self: XPathToken, context: ta.ContextType = None) \
        -> ta.OneArithmeticOrEmpty:
    op1: ta.NumericType | None
    op2: ta.NumericType | None
    op1, op2 = self.get_operands(context, cls=NumericProxy)
    if op1 is None:
        return []
    elif op2 is None:
        raise self.error('XPTY0004', '2nd operand is an empty sequence')
    elif op2 == 0 and isinstance(op2, float):
        return math.nan
    elif math.isinf(op2) and not math.isinf(op1) and op1 != 0:
        return op1 if self.parser.version != '1.0' else math.nan

    try:
        if isinstance(op1, int) and isinstance(op2, int):
            return op1 % op2 if op1 * op2 >= 0 else -(abs(op1) % op2)
        return op1 % op2  # type: ignore[operator]
    except TypeError as err:
        raise self.error('FORG0006', err) from None
    except (ZeroDivisionError, decimal.InvalidOperation):
        raise self.error('FOAR0001') from None


# Resolve the intrinsic ambiguity of some infix operators
@method('or')
@method('and')
@method('div')
@method('mod')
def nud__disambiguation_of_infix_operators(self: XPathToken) -> NameToken:
    return self.as_name()


###
# Union expressions
@method('|', bp=50)
def led__union_operator(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol in ('|', 'union'):
        left.concatenated = True
    self[:] = left, self.parser.expression(rbp=50)
    return self


@method('|')
def select__union_operator(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[XPathNode]:
    if context is None:
        raise self.missing_context()

    results = {item for k in range(2) for item in self[k].select(copy(context))}
    if any(not isinstance(x, XPathNode) for x in results):
        raise self.error('XPTY0004', 'only XPath nodes are allowed')
    elif self.concatenated:
        yield from cast(set[XPathNode], results)
    else:
        yield from cast(list[XPathNode], sorted(results, key=node_position))


###
# Path expressions
@method('//', bp=75)
def nud__descendant_path(self: XPathToken) -> XPathToken:
    if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
        self.parser.expected_next(*self.parser.PATH_STEP_SYMBOLS)

    self[:] = self.parser.expression(75),
    return self


@method('/', bp=75)
def nud__child_path(self: XPathToken) -> XPathToken:
    if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
        try:
            self.parser.expected_next(*self.parser.PATH_STEP_SYMBOLS)
        except SyntaxError:
            return self

    self[:] = self.parser.expression(75),
    return self


@method('//')
@method('/')
def led__child_or_descendant_path(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol in ('/', '//', ':', '[', '$'):
        pass
    elif left.label not in self.parser.PATH_STEP_LABELS and \
            left.symbol not in self.parser.PATH_STEP_SYMBOLS:
        raise self.wrong_syntax()

    if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
        self.parser.expected_next(*self.parser.PATH_STEP_SYMBOLS)

    self[:] = left, self.parser.expression(75)
    return self


@method('/')
def select__child_path(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    """
    Child path expression. Selects child:: axis as default (when bind to '*' or '(name)').
    """
    if context is None:
        raise self.missing_context()
    elif not self:
        if isinstance(context.root, DocumentNode):
            yield context.root
    elif len(self) == 1:
        if isinstance(context.document, DocumentNode):
            context.item = context.document
        elif context.root is None or isinstance(context.root.parent, ElementNode):
            return  # No root or a rooted subtree -> document root produce []
        else:
            context.item = context.root  # A fragment or a schema node
        yield from self[0].select(context)
    else:
        items: set[ta.ItemType] = set()
        for _ in self[0].select_with_focus(context):
            if not isinstance(context.item, XPathNode):
                msg = f"Intermediate step contains an atomic value {context.item!r}"
                raise self.error('XPTY0019', msg)

            for result in self[1].select(context):
                if not isinstance(result, XPathNode):
                    yield result
                elif result in items:
                    pass
                elif isinstance(result, ElementNode):
                    if result.value not in items:
                        items.add(result)
                        yield result
                else:
                    items.add(result)
                    yield result


@method('//')
def select__descendant_path(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    """Operator '//' is a short equivalent to /descendant-or-self::node()/"""
    if context is None:
        raise self.missing_context()
    elif len(self) == 2:
        items: set[ta.ItemType] = set()
        for _ in self[0].select_with_focus(context):
            if not isinstance(context.item, XPathNode):
                raise self.error('XPTY0019')

            for _ in context.iter_descendants():
                for result in self[1].select(context):
                    if not isinstance(result, XPathNode):
                        yield result
                    elif result in items:
                        pass
                    elif isinstance(result, ElementNode):
                        if result.value not in items:
                            items.add(result)
                            yield result
                    else:
                        items.add(result)
                        yield result

    else:
        if isinstance(context.document, DocumentNode):
            context.item = context.document
        elif context.root is None or isinstance(context.root.parent, ElementNode):
            return  # No root or a rooted subtree -> document root produce []
        else:
            context.item = context.root  # A fragment or a schema node

        items = set()
        for _ in context.iter_descendants():
            for result in self[0].select(context):
                if not isinstance(result, XPathNode):
                    items.add(result)
                elif result in items:
                    pass
                elif isinstance(result, ElementNode):
                    if result.value not in items:
                        items.add(result)
                else:
                    items.add(result)

        yield from sorted(items, key=node_position)


###
# Predicate filters
@method('[', bp=80)
def led__predicate(self: XPathToken, left: XPathToken) -> XPathToken:
    self[:] = left, self.parser.expression()
    self.parser.advance(']')
    return self


@method('[')
def select__predicate(self: XPathToken, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
    if context is None:
        raise self.missing_context()

    for _ in self[0].select_with_focus(context):
        if (self[1].label in ('axis', 'kind test') or self[1].symbol == '..') \
                and not isinstance(context.item, XPathNode):
            raise self.error('XPTY0020')

        predicate = list(self[1].select(copy(context)))

        if len(predicate) == 1 and isinstance(predicate[0], NumericProxy):
            if context.position == predicate[0]:
                yield context.item
        elif self.boolean_value(predicate):
            yield context.item


###
# Parenthesized expressions
@method('(', bp=100)
def nud__parenthesized_expr(self: XPathToken) -> XPathToken:
    self[:] = self.parser.expression(),
    self.parser.advance(')')
    return self


@method('(')
def evaluate__parenthesized_expr(self: XPathToken, context: ta.ContextType = None) \
        -> ta.ValueType:
    return self[0].evaluate(context)


@method('(')
def select__parenthesized_expr(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    return self[0].select(context)

# XPath 1.0 definitions continue into module xpath1_functions



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from .xpath30 import XPath30Parser
from .xpath31 import XPath31Parser

XPath3Parser = XPath31Parser

__all__ = ['XPath30Parser', 'XPath31Parser', 'XPath3Parser']
